"""Tests for blocked channel senders in backup/restore."""

from types import SimpleNamespace
from unittest.mock import MagicMock, patch

from meshy.backup import BackupManager
from meshy.models import Channel


def _make_win(storage):
    win = MagicMock()
    win.storage = storage
    win._storage = storage
    win.channels = []
    win.contacts = []
    win.self_info = {}
    win.device_info = SimpleNamespace(
        name="",
        public_key="",
        radio_freq=0,
        radio_bw=0,
        radio_sf=0,
        radio_cr=0,
        tx_power=0,
        latitude=0,
        longitude=0,
    )
    win._device_info = win.device_info
    return win


class TestBackupBlockedSenders:
    def test_export_includes_blocked_senders(self, storage):
        win = _make_win(storage)
        win.channels = [Channel(index=0, name="General", psk=bytes(16))]
        storage.save_channel(win.channels[0])
        storage.block_channel_sender("Spammer")
        storage.block_channel_sender("Troll")

        mgr = BackupManager()
        backup = mgr.export_backup(win, sections={"messages"})

        assert "blocked_channel_senders" in backup
        assert sorted(backup["blocked_channel_senders"]) == ["Spammer", "Troll"]

    def test_export_omits_key_when_no_blocked(self, storage):
        win = _make_win(storage)
        win.channels = []

        mgr = BackupManager()
        backup = mgr.export_backup(win, sections={"messages"})

        assert "blocked_channel_senders" not in backup

    @patch("meshy.backup.GLib")
    def test_import_restores_blocked_senders(self, _mock_glib, storage):
        win = _make_win(storage)
        backup = {"blocked_channel_senders": ["Alice", "Bob"]}

        mgr = BackupManager()
        mgr._do_import_data(win, backup, {"messages"})

        assert storage.get_blocked_channel_senders() == {"Alice", "Bob"}

    @patch("meshy.backup.GLib")
    def test_import_updates_frame_handler_cache(self, _mock_glib, storage):
        win = _make_win(storage)
        backup = {"blocked_channel_senders": ["Spammer"]}

        mgr = BackupManager()
        mgr._do_import_data(win, backup, {"messages"})

        win._frame_handler.update_blocked_senders.assert_called_once()

    @patch("meshy.backup.GLib")
    def test_import_without_blocked_skips_cache_update(self, _mock_glib, storage):
        win = _make_win(storage)
        backup = {}

        mgr = BackupManager()
        mgr._do_import_data(win, backup, {"messages"})

        win._frame_handler.update_blocked_senders.assert_not_called()

    @patch("meshy.backup.GLib")
    def test_round_trip(self, _mock_glib, storage):
        win = _make_win(storage)
        win.channels = [Channel(index=0, name="General", psk=bytes(16))]
        storage.save_channel(win.channels[0])
        storage.block_channel_sender("Spammer")
        storage.block_channel_sender("Troll")

        mgr = BackupManager()
        backup = mgr.export_backup(win, sections={"messages"})

        storage.unblock_channel_sender("Spammer")
        storage.unblock_channel_sender("Troll")
        assert storage.get_blocked_channel_senders() == set()

        mgr._do_import_data(win, backup, {"messages"})

        assert storage.get_blocked_channel_senders() == {"Spammer", "Troll"}

    def test_available_sections_includes_messages_for_blocked(self):
        backup = {"blocked_channel_senders": ["Spammer"]}
        sections = BackupManager.available_sections(backup)
        assert "messages" in sections


class TestBackupRetention:
    def test_export_includes_retention_levels(self, storage):
        win = _make_win(storage)
        win.channels = [Channel(index=0, name="General", psk=bytes(16), retention_days=30)]
        storage.save_channel(win.channels[0])

        mgr = BackupManager()
        backup = mgr.export_backup(win, sections={"channels"})

        assert "retention_levels" in backup
        assert backup["retention_levels"] == {"0": 30}

    def test_export_omits_default_retention(self, storage):
        win = _make_win(storage)
        win.channels = [Channel(index=0, name="General", psk=bytes(16), retention_days=0)]
        storage.save_channel(win.channels[0])

        mgr = BackupManager()
        backup = mgr.export_backup(win, sections={"channels"})

        assert "retention_levels" not in backup

    @patch("meshy.backup.GLib")
    def test_import_restores_retention(self, _mock_glib, storage):
        win = _make_win(storage)
        storage.save_channel(Channel(index=0, name="General", psk=bytes(16)))
        backup = {"retention_levels": {"0": 7}}

        mgr = BackupManager()
        mgr._do_import_data(win, backup, {"channels"})

        ch = storage.get_channel(0)
        assert ch.retention_days == 7

    def test_available_sections_includes_channels_for_retention(self):
        backup = {"retention_levels": {"0": 30}}
        sections = BackupManager.available_sections(backup)
        assert "channels" in sections



"""Tests for MessageController routing logic."""

from unittest.mock import MagicMock, patch

from meshy.message_controller import (
    FLOOD_RETRIES,
    PATH_RETRIES,
    MessageController,
)
from meshy.models import Contact, ContactType
from meshy.storage import Storage


def _make_contact(smart=True, force_flood=False, path_length=-1, path=b""):
    c = Contact(
        public_key=bytes(range(32)),
        name="Test",
        type=ContactType.CHAT,
        smart_route_selection=smart,
        force_flood=force_flood,
        path_length=path_length,
        path=path,
    )
    return c


def _make_win(storage):
    win = MagicMock()
    win._storage = storage
    return win


def _make_ctrl(storage=None):
    if storage is None:
        storage = Storage(db_path=":memory:")
    win = _make_win(storage)
    ctrl = MessageController(win)
    return ctrl, storage


# ─── _determine_strategy() ───────────────────────────────────


class TestDetermineStrategy:
    def test_override_direct(self):
        c = _make_contact(smart=True)
        assert MessageController._determine_strategy(c, "direct") == "override_direct"

    def test_override_flood(self):
        c = _make_contact(smart=True)
        assert MessageController._determine_strategy(c, "flood") == "override_flood"

    def test_force_flood(self):
        c = _make_contact(smart=False, force_flood=True)
        assert MessageController._determine_strategy(c, None) == "force_flood"

    def test_smart(self):
        c = _make_contact(smart=True)
        assert MessageController._determine_strategy(c, None) == "smart"

    def test_traditional(self):
        c = _make_contact(smart=False)
        assert MessageController._determine_strategy(c, None) == "traditional"

    def test_override_takes_precedence_over_force_flood(self):
        c = _make_contact(smart=True, force_flood=True)
        assert MessageController._determine_strategy(c, "direct") == "override_direct"


# ─── max_retries_for() ────────────────────────────────────────


class TestMaxRetries:
    def setup_method(self):
        self.ctrl, self.storage = _make_ctrl()

    def teardown_method(self):
        self.storage.close()

    def test_override_direct_returns_1(self):
        ctx = {"strategy": "override_direct"}
        assert self.ctrl.max_retries_for(ctx) == 1

    def test_override_flood_returns_flood_retries(self):
        ctx = {"strategy": "override_flood"}
        assert self.ctrl.max_retries_for(ctx) == FLOOD_RETRIES

    def test_force_flood_returns_flood_retries(self):
        ctx = {"strategy": "force_flood"}
        assert self.ctrl.max_retries_for(ctx) == FLOOD_RETRIES

    def test_smart_no_history_returns_flood_retries(self):
        contact = _make_contact(smart=True)
        ctx = {"strategy": "smart", "contact": contact}
        assert self.ctrl.max_retries_for(ctx) == FLOOD_RETRIES

    def test_smart_with_history_returns_path_plus_one(self):
        contact = _make_contact(smart=True)
        ctx = {"strategy": "smart", "contact": contact}
        self.storage.upsert_path_record(
            contact.public_key_hex,
            "aabb",
            1,
            2,
            success=True,
            trip_ms=100,
        )
        assert self.ctrl.max_retries_for(ctx) == PATH_RETRIES + 1

    def test_smart_retries_update_live_after_path_added(self):
        contact = _make_contact(smart=True)
        ctx = {"strategy": "smart", "contact": contact}
        assert self.ctrl.max_retries_for(ctx) == FLOOD_RETRIES
        self.storage.upsert_path_record(
            contact.public_key_hex,
            "aabb",
            1,
            2,
            success=True,
            trip_ms=100,
        )
        assert self.ctrl.max_retries_for(ctx) == PATH_RETRIES + 1

    def test_smart_retries_drop_after_weight_zero(self):
        contact = _make_contact(smart=True)
        ctx = {"strategy": "smart", "contact": contact}
        self.storage.upsert_path_record(
            contact.public_key_hex,
            "aabb",
            1,
            2,
            success=True,
        )
        assert self.ctrl.max_retries_for(ctx) == PATH_RETRIES + 1
        for _ in range(20):
            self.storage.upsert_path_record(
                contact.public_key_hex,
                "aabb",
                1,
                2,
                success=False,
                weight_increment=-0.2,
            )
        assert self.ctrl.max_retries_for(ctx) == FLOOD_RETRIES

    def test_traditional_with_path_returns_path_plus_one(self):
        ctx = {"strategy": "traditional", "started_with_path": True}
        assert self.ctrl.max_retries_for(ctx) == PATH_RETRIES + 1

    def test_traditional_flood_returns_flood_retries(self):
        ctx = {"strategy": "traditional", "started_with_path": False}
        assert self.ctrl.max_retries_for(ctx) == FLOOD_RETRIES


# ─── _apply_smart_route() ────────────────────────────────────


class TestApplySmartRoute:
    def setup_method(self):
        self.ctrl, self.storage = _make_ctrl()

    def teardown_method(self):
        self.storage.close()

    def test_no_history_resets_to_flood(self):
        contact = _make_contact(smart=True, path_length=2, path=b"\xaa\xbb")
        ctx = {"contact": contact, "attempt": 0}
        self.ctrl._apply_smart_route(ctx, contact)
        assert contact.path_length == -1
        assert contact.path == b""

    def test_attempt_past_path_retries_resets_to_flood(self):
        contact = _make_contact(smart=True, path_length=2, path=b"\xaa\xbb")
        self.storage.upsert_path_record(
            contact.public_key_hex,
            "aabb",
            1,
            2,
            success=True,
            trip_ms=100,
        )
        ctx = {"contact": contact, "attempt": PATH_RETRIES}
        self.ctrl._apply_smart_route(ctx, contact)
        assert contact.path_length == -1

    def test_selects_best_path(self):
        contact = _make_contact(smart=True)
        path_hex = "cc" * 2
        self.storage.upsert_path_record(
            contact.public_key_hex,
            path_hex,
            1,
            2,
            success=True,
            trip_ms=100,
        )
        ctx = {"contact": contact, "attempt": 0}
        self.ctrl._apply_smart_route(ctx, contact)
        self.ctrl._win.set_contact_path.assert_called_once()
        args = self.ctrl._win.set_contact_path.call_args[0]
        assert args[0] is contact

    def test_excludes_tried_paths(self):
        contact = _make_contact(smart=True)
        for i in range(3):
            path_hex = f"{i:02x}" * 2
            for _ in range(5):
                self.storage.upsert_path_record(
                    contact.public_key_hex,
                    path_hex,
                    1,
                    2,
                    success=True,
                    trip_ms=100 + i * 100,
                )

        ctx = {"contact": contact, "attempt": 0, "paths_tried": ["0000"]}
        self.ctrl._apply_smart_route(ctx, contact)
        set_path_args = self.ctrl._win.set_contact_path.call_args[0]
        set_path_bytes = set_path_args[2]
        assert set_path_bytes != bytes.fromhex("0000")

    def test_resets_tried_when_all_exhausted(self):
        contact = _make_contact(smart=True)
        path_hex = "aabb"
        self.storage.upsert_path_record(
            contact.public_key_hex,
            path_hex,
            1,
            2,
            success=True,
            trip_ms=100,
        )
        ctx = {"contact": contact, "attempt": 1, "paths_tried": [path_hex]}
        self.ctrl._apply_smart_route(ctx, contact)
        assert ctx["paths_tried"] == []
        self.ctrl._win.set_contact_path.assert_called_once()

    def test_does_not_reset_path_when_already_set(self):
        contact = _make_contact(smart=True, path_length=2, path=b"\xaa\xbb")
        self.storage.upsert_path_record(
            contact.public_key_hex,
            "aabb",
            1,
            2,
            success=True,
            trip_ms=100,
        )
        ctx = {"contact": contact, "attempt": 0, "paths_tried": ["aabb"]}
        self.ctrl._apply_smart_route(ctx, contact)
        assert contact.path_length == 2
        self.ctrl._win.set_contact_path.assert_not_called()


# ─── _send_message_attempt() pre-send refresh ────────────────


class TestPreSendRefresh:
    def setup_method(self):
        self.ctrl, self.storage = _make_ctrl()

    def teardown_method(self):
        self.storage.close()

    def test_first_smart_attempt_triggers_refresh(self):
        contact = _make_contact(smart=True)
        msg_id = "test-msg-1"
        self.ctrl.retry_context[msg_id] = {
            "contact": contact,
            "strategy": "smart",
            "attempt": 0,
        }
        self.ctrl._send_message_attempt(msg_id)
        self.ctrl._win._frame_handler.do_contacts_refetch.assert_called_once()
        assert self.ctrl.retry_context[msg_id]["_refreshed"] is True

    @patch.object(MessageController, "_do_send_message", return_value=False)
    def test_refresh_only_once(self, mock_send):
        contact = _make_contact(smart=True)
        msg_id = "test-msg-2"
        self.ctrl.retry_context[msg_id] = {
            "contact": contact,
            "strategy": "smart",
            "attempt": 0,
            "_refreshed": True,
        }
        self.ctrl._send_message_attempt(msg_id)
        self.ctrl._win._frame_handler.do_contacts_refetch.assert_not_called()
        mock_send.assert_called_once_with(msg_id)

    @patch.object(MessageController, "_do_send_message", return_value=False)
    def test_non_smart_skips_refresh(self, mock_send):
        contact = _make_contact(smart=False)
        msg_id = "test-msg-3"
        self.ctrl.retry_context[msg_id] = {
            "contact": contact,
            "strategy": "traditional",
            "attempt": 0,
        }
        self.ctrl._send_message_attempt(msg_id)
        self.ctrl._win._frame_handler.do_contacts_refetch.assert_not_called()
        mock_send.assert_called_once_with(msg_id)

    @patch.object(MessageController, "_do_send_message", return_value=False)
    def test_retry_attempt_skips_refresh(self, mock_send):
        contact = _make_contact(smart=True)
        msg_id = "test-msg-4"
        self.ctrl.retry_context[msg_id] = {
            "contact": contact,
            "strategy": "smart",
            "attempt": 2,
        }
        self.ctrl._send_message_attempt(msg_id)
        self.ctrl._win._frame_handler.do_contacts_refetch.assert_not_called()
        mock_send.assert_called_once_with(msg_id)



# Copyright 2026, Jiri Eischmann <jiri@eischmann.cz> and the meshy contributors
# SPDX-License-Identifier: GPL-3.0-or-later

"""Backup and restore for MeshCore device configuration and messages."""

import glob
import json
import logging
import os
import uuid
from datetime import datetime

from gi.repository import GLib
from meshy.models import (
    ChannelMessage,
    Contact,
    Message,
    MessageStatus,
    NotificationLevel,
)
from meshy.protocol import (
    build_add_update_contact,
    build_device_query,
    build_set_auto_add_config,
    build_set_channel,
)

log = logging.getLogger(__name__)


class BackupManager:
    """Stateless helper for building and applying backup dicts."""

    ALL_SECTIONS = {
        "device_settings",
        "device_identity",
        "contacts",
        "channels",
        "messages",
        "regions",
        "discovered_contacts",
        "path_history",
        "local_settings",
    }

    def export_backup(self, win, sections: set[str] | None = None) -> dict:
        if sections is None:
            sections = self.ALL_SECTIONS

        si = win.self_info
        di = win.device_info

        backup = {
            "name": di.name or si.get("name", ""),
            "public_key": di.public_key or si.get("public_key", ""),
        }

        if "device_settings" in sections:
            backup["radio_settings"] = {
                "frequency": int(di.radio_freq * 1000)
                if di.radio_freq
                else si.get("radio_freq", 0) * 1000,
                "bandwidth": int(di.radio_bw * 1000)
                if di.radio_bw
                else si.get("radio_bw", 0) * 1000,
                "spreading_factor": di.radio_sf or si.get("radio_sf", 0),
                "coding_rate": di.radio_cr or si.get("radio_cr", 0),
                "tx_power": di.tx_power or si.get("tx_power", 0),
            }
            backup["position_settings"] = {
                "latitude": str(si.get("adv_lat", 0.0)),
                "longitude": str(si.get("adv_lon", 0.0)),
            }
            backup["other_settings"] = {
                "manual_add_contacts": 1 if si.get("manual_add_contacts") else 0,
                "advert_location_policy": si.get("adv_loc_policy", 0),
            }
            backup["auto_add_settings"] = win.get_auto_add_config()

        if "channels" in sections:
            channels_list = []
            for ch in win.channels:
                if not ch.is_empty:
                    channels_list.append(
                        {
                            "name": ch.name,
                            "secret": ch.psk_hex,
                        }
                    )
            if channels_list:
                backup["channels"] = channels_list

            retention = {}
            for ch in win.channels:
                if not ch.is_empty and ch.retention_days != 0:
                    retention[str(ch.index)] = ch.retention_days
            if retention:
                backup["retention_levels"] = retention

        if "contacts" in sections:
            contacts_list = []
            for c in win.contacts:
                contacts_list.append(
                    {
                        "type": c.type,
                        "name": c.name,
                        "public_key": c.public_key_hex,
                        "flags": c.flags,
                        "latitude": str(c.latitude) if c.latitude else "0.0",
                        "longitude": str(c.longitude) if c.longitude else "0.0",
                        "last_advert": int(c.last_seen.timestamp()) if c.last_seen else 0,
                        "last_modified": int(c.device_lastmod.timestamp())
                        if c.device_lastmod
                        else 0,
                        "out_path": ",".join(c.path_hops) if c.path_hops else None,
                    }
                )
            if contacts_list:
                backup["contacts"] = contacts_list

        if win.storage:
            if "messages" in sections:
                dm_messages = {}
                for c in win.contacts:
                    msgs = win.storage.get_messages(c.public_key_hex, limit=10000)
                    if msgs:
                        dm_messages[c.public_key_hex] = [
                            {
                                "text": m.text,
                                "timestamp": m.timestamp.timestamp(),
                                "is_outgoing": m.is_outgoing,
                                "sender_key_hex": m.sender_key_hex,
                                "room_sender_name": m.room_sender_name,
                            }
                            for m in msgs
                        ]
                if dm_messages:
                    backup["messages"] = dm_messages

                ch_messages = {}
                for ch in win.channels:
                    if ch.is_empty:
                        continue
                    msgs = win.storage.get_channel_messages(ch.index, limit=10000)
                    if msgs:
                        ch_messages[str(ch.index)] = [
                            {
                                "sender_name": m.sender_name,
                                "text": m.text,
                                "timestamp": m.timestamp.timestamp(),
                                "is_outgoing": m.is_outgoing,
                            }
                            for m in msgs
                        ]
                if ch_messages:
                    backup["channel_messages"] = ch_messages

                passwords = {}
                for c in win.contacts:
                    pw = win.storage.get_room_password(c.public_key_hex)
                    if pw:
                        passwords[c.public_key_hex] = pw
                if passwords:
                    backup["room_passwords"] = passwords

                notif = {}
                for ch in win.channels:
                    if not ch.is_empty and ch.notification_level != NotificationLevel.DEFAULT:
                        notif[str(ch.index)] = ch.notification_level.value
                if notif:
                    backup["notification_levels"] = notif

                blocked = list(win.storage.get_blocked_channel_senders())
                if blocked:
                    backup["blocked_channel_senders"] = sorted(blocked)

            if "regions" in sections:
                regions = win.storage.get_regions()
                if regions:
                    backup["regions"] = regions

            if "discovered_contacts" in sections:
                discovered = [
                    {
                        "public_key": c.public_key_hex,
                        "name": c.name,
                        "type": c.type,
                        "flags": c.flags,
                        "latitude": c.latitude,
                        "longitude": c.longitude,
                        "last_seen": c.last_seen.timestamp() if c.last_seen else None,
                        "received_at": c.received_at.timestamp() if c.received_at else None,
                    }
                    for c in win.storage.get_discovered_contacts()
                ]
                if discovered:
                    backup["discovered_contacts"] = discovered

            if "path_history" in sections:
                all_history = {}
                for c in win.contacts:
                    entries = win.storage.get_path_history(c.public_key_hex)
                    if entries:
                        all_history[c.public_key_hex] = [
                            {
                                "path_bytes": e["path_bytes"],
                                "path_hash_size": e["path_hash_size"],
                                "hop_count": e["hop_count"],
                                "success_count": e["success_count"],
                                "failure_count": e["failure_count"],
                                "last_trip_ms": e["last_trip_ms"],
                                "route_weight": e["route_weight"],
                                "last_success_at": e["last_success_at"],
                            }
                            for e in entries
                        ]
                if all_history:
                    backup["path_history"] = all_history

            if "local_settings" in sections:
                local = {}
                scope = win.storage.get_default_scope()
                if scope:
                    local["default_scope"] = scope
                if local:
                    backup["local_settings"] = local

        return backup

    _BACKUP_KEYS_FOR_SECTION = {
        "device_settings": [
            "name",
            "radio_settings",
            "position_settings",
            "other_settings",
            "auto_add_settings",
        ],
        "device_identity": ["private_key"],
        "contacts": ["contacts"],
        "channels": ["channels", "retention_levels"],
        "messages": [
            "messages",
            "channel_messages",
            "room_passwords",
            "notification_levels",
            "blocked_channel_senders",
        ],
        "regions": ["regions"],
        "discovered_contacts": ["discovered_contacts"],
        "path_history": ["path_history"],
        "local_settings": ["local_settings"],
    }

    @classmethod
    def available_sections(cls, backup: dict) -> set[str]:
        result = set()
        for section, keys in cls._BACKUP_KEYS_FOR_SECTION.items():
            for key in keys:
                val = backup.get(key)
                if val:
                    result.add(section)
                    break
        return result

    def import_backup(self, win, backup: dict, sections: set[str] | None = None, on_complete=None):
        if sections is None:
            sections = self.ALL_SECTIONS

        key_hex = backup.get("private_key", "")
        if "device_identity" in sections and key_hex and len(key_hex) == 128:
            try:
                key_bytes = bytes.fromhex(key_hex)
            except ValueError:
                key_bytes = None

            if key_bytes:

                def _after_key():
                    result = self._do_import_data(win, backup, sections)
                    if on_complete:
                        on_complete(result)

                def _on_key_error(err):
                    log.warning(
                        "Private key import failed: %s",
                        err.get("message", "") if err else "unknown",
                    )
                    _after_key()

                win._frame_handler.expect_ok(
                    on_ok=_after_key,
                    on_error=_on_key_error,
                )
                win.import_private_key(key_bytes)
                return 0, 0

        result = self._do_import_data(win, backup, sections)
        if on_complete:
            on_complete(result)
        return result

    def _do_import_data(self, win, backup: dict, sections: set[str]):
        imported_contacts = 0
        imported_channels = 0

        settings_cmds = []
        if "device_settings" in sections:
            name = backup.get("name", "")
            if name:
                settings_cmds.append(lambda: win.set_advert_name(name))

            radio = backup.get("radio_settings", {})
            if radio:
                freq = radio.get("frequency", 0)
                bw = radio.get("bandwidth", 0)
                sf = radio.get("spreading_factor", 0)
                cr = radio.get("coding_rate", 0)
                tx = radio.get("tx_power", 0)
                if freq and bw and sf and cr:
                    settings_cmds.append(lambda: win.set_radio_params(freq, bw, sf, cr))
                if tx:
                    settings_cmds.append(lambda: win.set_tx_power(tx))

            pos = backup.get("position_settings", {})
            if pos:
                try:
                    lat = float(pos.get("latitude", 0))
                    lon = float(pos.get("longitude", 0))
                    if abs(lat) > 1e-6 or abs(lon) > 1e-6:
                        settings_cmds.append(lambda: win.set_advert_latlon(lat, lon))
                except (ValueError, TypeError):
                    pass

            other = backup.get("other_settings", {})
            if other:
                loc_policy = other.get("advert_location_policy", None)
                if loc_policy is not None:
                    settings_cmds.append(lambda: win.set_advert_loc_policy(loc_policy))

            auto = backup.get("auto_add_settings", {})
            if auto:
                frame = build_set_auto_add_config(
                    chat=auto.get("auto_add_chat", False),
                    repeater=auto.get("auto_add_repeater", False),
                    room=auto.get("auto_add_room_server", False),
                    sensor=auto.get("auto_add_sensor", False),
                    overwrite_oldest=auto.get("overwrite_oldest", False),
                    max_hops=auto.get("max_hops", 0),
                )
                settings_cmds.append(lambda: win.send_frame(frame))

        for i, cmd in enumerate(settings_cmds):
            GLib.timeout_add(i * 300, lambda c=cmd: c() or False)
        settings_delay = len(settings_cmds) * 300

        if "device_settings" in sections:
            di = win._device_info
            name = backup.get("name", "")
            if name:
                di.name = name
            radio = backup.get("radio_settings", {})
            if radio:
                freq = radio.get("frequency", 0)
                bw = radio.get("bandwidth", 0)
                sf = radio.get("spreading_factor", 0)
                cr = radio.get("coding_rate", 0)
                tx = radio.get("tx_power", 0)
                if freq and bw and sf and cr:
                    di.radio_freq = freq
                    di.radio_bw = bw
                    di.radio_sf = sf
                    di.radio_cr = cr
                if tx:
                    di.tx_power = tx
            pos = backup.get("position_settings", {})
            if pos:
                try:
                    lat_v = float(pos.get("latitude", 0))
                    lon_v = float(pos.get("longitude", 0))
                    if abs(lat_v) > 1e-6 or abs(lon_v) > 1e-6:
                        di.latitude = lat_v
                        di.longitude = lon_v
                except (ValueError, TypeError):
                    pass
            win._settings_view.update_device_info(di)

        device_cmds = []
        if "channels" in sections:
            channels = backup.get("channels", [])
            used = {c.index for c in win.channels if not c.is_empty}
            for ch_data in channels:
                ch_name = ch_data.get("name", "")
                secret_hex = ch_data.get("secret", "")
                if ch_name and secret_hex and len(secret_hex) == 32:
                    try:
                        psk = bytes.fromhex(secret_hex)
                        already = any(
                            c.psk_hex == secret_hex for c in win.channels if not c.is_empty
                        )
                        if already:
                            continue
                        slot = None
                        for idx in range(win.device_info.max_channels or 8):
                            if idx not in used:
                                slot = idx
                                break
                        if slot is not None:
                            device_cmds.append(
                                lambda s=slot, n=ch_name, p=psk: win.send_frame(
                                    build_set_channel(s, n, p)
                                )
                            )
                            imported_channels += 1
                            used.add(slot)
                    except ValueError:
                        pass

            if win.storage:
                for ch_idx_str, days in backup.get("retention_levels", {}).items():
                    win.storage.set_channel_retention_days(int(ch_idx_str), days)

        if "contacts" in sections:
            contacts = backup.get("contacts", [])
            contacts_by_key = {c.public_key_hex: c for c in win.contacts}
            for c_data in contacts:
                pub_key_hex = c_data.get("public_key", "")
                if not pub_key_hex or len(pub_key_hex) != 64:
                    continue
                if pub_key_hex in contacts_by_key:
                    continue
                try:
                    pub_key = bytes.fromhex(pub_key_hex)
                    c_name = c_data.get("name", "Unknown")
                    contact_type = c_data.get("type", 1)
                    flags = c_data.get("flags", 0)
                    lat = float(c_data.get("latitude", 0))
                    lon = float(c_data.get("longitude", 0))
                    last_mod = c_data.get("last_modified")
                    device_cmds.append(
                        lambda pk=pub_key,
                        ct=contact_type,
                        fl=flags,
                        n=c_name,
                        la=lat,
                        lo=lon,
                        lm=last_mod: win.send_frame(
                            build_add_update_contact(
                                pub_key=pk,
                                contact_type=ct,
                                flags=fl,
                                name=n,
                                lat=la if abs(la) > 1e-6 else None,
                                lon=lo if abs(lo) > 1e-6 else None,
                                last_modified=lm,
                            )
                        )
                    )
                    imported_contacts += 1
                except (ValueError, TypeError):
                    pass

        for i, cmd in enumerate(device_cmds):
            GLib.timeout_add(settings_delay + i * 300, lambda c=cmd: c() or False)

        if "messages" in sections and win.storage:
            with win.storage.batch():
                for pub_key_hex, msgs in backup.get("messages", {}).items():
                    for m in msgs:
                        ts = m.get("timestamp", 0)
                        text = m.get("text", "")
                        if win.storage.message_exists(pub_key_hex, ts, text):
                            continue
                        message = Message(
                            sender_key=bytes.fromhex(m.get("sender_key_hex", pub_key_hex)),
                            text=text,
                            timestamp=datetime.fromtimestamp(ts),
                            is_outgoing=m.get("is_outgoing", False),
                            status=MessageStatus.DELIVERED,
                            message_id=str(uuid.uuid4()),
                            room_sender_name=m.get("room_sender_name", ""),
                        )
                        win.storage.save_message(pub_key_hex, message)

                for ch_idx_str, msgs in backup.get("channel_messages", {}).items():
                    ch_idx = int(ch_idx_str)
                    for m in msgs:
                        msg = ChannelMessage(
                            channel_index=ch_idx,
                            sender_name=m.get("sender_name", ""),
                            text=m.get("text", ""),
                            timestamp=datetime.fromtimestamp(m.get("timestamp", 0)),
                            is_outgoing=m.get("is_outgoing", False),
                        )
                        win.storage.save_channel_message(msg)

                for pub_key_hex, pw in backup.get("room_passwords", {}).items():
                    win.storage.save_room_password(pub_key_hex, pw)

                for name in backup.get("blocked_channel_senders", []):
                    win.storage.block_channel_sender(name)

        if win.storage:
            if "regions" in sections:
                for region_name in backup.get("regions", []):
                    win.storage.add_region(region_name)

            if "discovered_contacts" in sections:
                for dc in backup.get("discovered_contacts", []):
                    pub_key_hex = dc.get("public_key", "")
                    if not pub_key_hex or len(pub_key_hex) != 64:
                        continue
                    contact = Contact(
                        public_key=bytes.fromhex(pub_key_hex),
                        name=dc.get("name", ""),
                        type=dc.get("type", 1),
                        flags=dc.get("flags", 0),
                        latitude=dc.get("latitude"),
                        longitude=dc.get("longitude"),
                        last_seen=datetime.fromtimestamp(dc["last_seen"])
                        if dc.get("last_seen")
                        else None,
                        received_at=datetime.fromtimestamp(dc["received_at"])
                        if dc.get("received_at")
                        else None,
                    )
                    win.storage.save_discovered_contact(contact)

            if "path_history" in sections:
                for contact_key, entries in backup.get("path_history", {}).items():
                    for e in entries:
                        win.storage.import_path_history_entry(
                            contact_key_hex=contact_key,
                            path_bytes=e["path_bytes"],
                            path_hash_size=e.get("path_hash_size", 1),
                            hop_count=e.get("hop_count", 0),
                            success_count=e.get("success_count", 0),
                            failure_count=e.get("failure_count", 0),
                            last_trip_ms=e.get("last_trip_ms"),
                            route_weight=e.get("route_weight", 2.5),
                            last_success_at=e.get("last_success_at"),
                        )

            if "local_settings" in sections:
                local = backup.get("local_settings", {})
                if "default_scope" in local:
                    win.storage.set_default_scope(local["default_scope"])

        if "messages" in sections and backup.get("blocked_channel_senders"):
            win._frame_handler.update_blocked_senders()

        total_delay = settings_delay + len(device_cmds) * 300 + 1000
        GLib.timeout_add(total_delay, lambda: win._frame_handler.send_get_contacts() or False)
        GLib.timeout_add(total_delay + 500, lambda: win._fetch_channels_sequential(0) or False)
        GLib.timeout_add(total_delay + 1000, lambda: win.send_frame(build_device_query()) or False)

        return imported_contacts, imported_channels


_DAY_MS = 24 * 60 * 60 * 1000
_MAX_AUTO_BACKUPS = 7


class AutoBackupManager:
    def __init__(self):
        self._timer_id = None
        self._backup = BackupManager()

    @staticmethod
    def _backups_dir() -> str:
        data_dir = os.path.join(
            os.environ.get("XDG_DATA_HOME", os.path.expanduser("~/.local/share")),
            "meshy",
            "backups",
        )
        os.makedirs(data_dir, exist_ok=True)
        return data_dir

    @staticmethod
    def _safe_address(device_address: str) -> str:
        return device_address.replace(":", "_").replace("/", "_")

    def run(self, win) -> bool:
        if not win.is_connected or not win.storage:
            return False
        try:
            from gi.repository import Gio

            settings = Gio.Settings.new("page.codeberg.sesivany.Meshy")
            sections = set(settings.get_strv("auto-backup-sections"))
            sections.discard("device_identity")
            backup = self._backup.export_backup(win, sections=sections or None)
            addr = self._safe_address(win._transport.connected_address or "unknown")
            date_str = datetime.now().strftime("%Y-%m-%d")
            filename = f"auto_{addr}_{date_str}.json"
            path = os.path.join(self._backups_dir(), filename)
            with open(path, "w", encoding="utf-8") as f:
                json.dump(backup, f, indent=2, ensure_ascii=False)
            log.info("Auto-backup saved: %s", path)
            self._prune(addr)
            return True
        except Exception as e:
            log.error("Auto-backup failed: %s", e)
            return False

    def schedule(self, win):
        self.cancel()
        self.run(win)

        def _daily_backup():
            self.run(win)
            return True

        self._timer_id = GLib.timeout_add(_DAY_MS, _daily_backup)

    def cancel(self):
        if self._timer_id is not None:
            GLib.source_remove(self._timer_id)
            self._timer_id = None

    def list_backups(self, device_address: str | None = None) -> list[dict]:
        backups_dir = self._backups_dir()
        if device_address:
            pattern = f"auto_{self._safe_address(device_address)}_*.json"
        else:
            pattern = "auto_*.json"
        files = sorted(glob.glob(os.path.join(backups_dir, pattern)), reverse=True)
        result = []
        for path in files:
            try:
                with open(path, encoding="utf-8") as f:
                    data = json.load(f)
                mtime = os.path.getmtime(path)
                mdt = datetime.fromtimestamp(mtime)
                result.append(
                    {
                        "path": path,
                        "filename": os.path.basename(path),
                        "name": data.get("name", ""),
                        "contacts": len(data.get("contacts", [])),
                        "channels": len(data.get("channels", [])),
                        "date": os.path.basename(path).rsplit("_", 1)[-1].replace(".json", ""),
                        "time": mdt.strftime("%H:%M"),
                        "mtime_dt": mdt,
                    }
                )
            except Exception:
                continue
        return result

    def _prune(self, safe_address: str):
        pattern = os.path.join(self._backups_dir(), f"auto_{safe_address}_*.json")
        files = sorted(glob.glob(pattern))
        while len(files) > _MAX_AUTO_BACKUPS:
            oldest = files.pop(0)
            try:
                os.remove(oldest)
                log.info("Pruned old auto-backup: %s", oldest)
            except OSError as e:
                log.warning("Failed to prune %s: %s", oldest, e)



from unittest.mock import patch

from meshy.models import Channel, Contact, MessageStatus


def _insert_msg(storage, contact_hex, text, ts, is_outgoing=False):
    storage._db.execute(
        """INSERT INTO messages
        (contact_key_hex, sender_key_hex, text, timestamp, is_outgoing, is_cli, status, message_id, room_sender_name)
        VALUES (?, ?, ?, ?, ?, 0, 0, NULL, '')""",
        (contact_hex, contact_hex, text, ts, int(is_outgoing)),
    )
    storage._db.commit()


CONTACT_HEX = "aa" * 32
OTHER_HEX = "bb" * 32


# ─── has_duplicate_message ─────────────────────────────────────


class TestHasDuplicateMessage:
    def test_empty_db_returns_false(self, storage):
        assert storage.has_duplicate_message(CONTACT_HEX, 1000.0, "hello") is False

    def test_layer1_exact_timestamp_match(self, storage):
        _insert_msg(storage, CONTACT_HEX, "hello", 1000.0)
        assert storage.has_duplicate_message(CONTACT_HEX, 1000.0, "hello") is True

    def test_layer1_different_text_not_duplicate(self, storage):
        _insert_msg(storage, CONTACT_HEX, "hello", 1000.0)
        assert storage.has_duplicate_message(CONTACT_HEX, 1000.0, "world") is False

    @patch("meshy.storage.time")
    def test_layer2_manual_resend_within_30min(self, mock_time, storage):
        ts = 1000.0
        mock_time.time.return_value = ts + 600  # 10 min later
        _insert_msg(storage, CONTACT_HEX, "hello", ts)
        assert storage.has_duplicate_message(CONTACT_HEX, ts + 60, "hello") is True

    @patch("meshy.storage.time")
    def test_layer2_not_duplicate_after_30min(self, mock_time, storage):
        ts = 1000.0
        mock_time.time.return_value = ts + 2000  # > 30 min
        _insert_msg(storage, CONTACT_HEX, "hello", ts)
        assert storage.has_duplicate_message(CONTACT_HEX, ts + 60, "hello") is False

    @patch("meshy.storage.time")
    def test_layer2_not_duplicate_after_our_reply(self, mock_time, storage):
        ts = 1000.0
        mock_time.time.return_value = ts + 300
        _insert_msg(storage, CONTACT_HEX, "hello", ts)
        _insert_msg(storage, CONTACT_HEX, "reply", ts + 100, is_outgoing=True)
        assert storage.has_duplicate_message(CONTACT_HEX, ts + 200, "hello") is False

    @patch("meshy.storage.time")
    def test_layer2_different_text_not_duplicate(self, mock_time, storage):
        ts = 1000.0
        mock_time.time.return_value = ts + 300
        _insert_msg(storage, CONTACT_HEX, "hello", ts)
        assert storage.has_duplicate_message(CONTACT_HEX, ts + 60, "world") is False


# ─── message_exists ────────────────────────────────────────────


class TestMessageExists:
    def test_within_5s_window(self, storage):
        _insert_msg(storage, CONTACT_HEX, "test", 1000.0)
        assert storage.message_exists(CONTACT_HEX, 1003.0, "test") is True

    def test_outside_5s_window(self, storage):
        _insert_msg(storage, CONTACT_HEX, "test", 1000.0)
        assert storage.message_exists(CONTACT_HEX, 1010.0, "test") is False

    def test_different_text(self, storage):
        _insert_msg(storage, CONTACT_HEX, "test", 1000.0)
        assert storage.message_exists(CONTACT_HEX, 1000.0, "other") is False


# ─── count_unread_messages ─────────────────────────────────────


class TestCountUnread:
    def test_all_unread_without_last_read(self, storage):
        storage.save_contact(Contact(public_key=bytes.fromhex(CONTACT_HEX), name="Alice"))
        _insert_msg(storage, CONTACT_HEX, "m1", 100.0)
        _insert_msg(storage, CONTACT_HEX, "m2", 200.0)
        _insert_msg(storage, CONTACT_HEX, "m3", 300.0, is_outgoing=True)
        assert storage.count_unread_messages(CONTACT_HEX) == 2

    def test_partial_unread_with_last_read(self, storage):
        storage.save_contact(Contact(public_key=bytes.fromhex(CONTACT_HEX), name="Alice"))
        _insert_msg(storage, CONTACT_HEX, "m1", 100.0)
        _insert_msg(storage, CONTACT_HEX, "m2", 200.0)
        _insert_msg(storage, CONTACT_HEX, "m3", 300.0)
        storage.set_contact_last_read_at(CONTACT_HEX, 150.0)
        assert storage.count_unread_messages(CONTACT_HEX) == 2

    def test_channel_unread(self, storage):
        storage.save_channel(Channel(index=0, name="General", psk=bytes(16)))
        storage._db.execute(
            "INSERT INTO channel_messages (channel_index, sender_name, text, timestamp, is_outgoing) VALUES (0, 'Bob', 'hi', 100.0, 0)",
        )
        storage._db.execute(
            "INSERT INTO channel_messages (channel_index, sender_name, text, timestamp, is_outgoing) VALUES (0, 'Bob', 'there', 200.0, 0)",
        )
        storage._db.commit()
        assert storage.count_unread_channel_messages(0) == 2
        storage.set_channel_last_read_at(0, 150.0)
        assert storage.count_unread_channel_messages(0) == 1


# ─── upsert_path_record / get_path_history ─────────────────────


class TestPathHistory:
    def test_insert_new_record(self, storage):
        storage.upsert_path_record(CONTACT_HEX, "aabb", 1, 2, success=True, trip_ms=500)
        history = storage.get_path_history(CONTACT_HEX)
        assert len(history) == 1
        rec = history[0]
        assert rec["path_bytes"] == "aabb"
        assert rec["success_count"] == 1
        assert rec["failure_count"] == 0
        assert rec["last_trip_ms"] == 500

    def test_update_increments_counts(self, storage):
        storage.upsert_path_record(CONTACT_HEX, "aabb", 1, 2, success=True, trip_ms=500)
        storage.upsert_path_record(CONTACT_HEX, "aabb", 1, 2, success=True, trip_ms=300)
        storage.upsert_path_record(CONTACT_HEX, "aabb", 1, 2, success=False)
        history = storage.get_path_history(CONTACT_HEX)
        assert len(history) == 1
        assert history[0]["success_count"] == 2
        assert history[0]["failure_count"] == 1

    def test_history_limit_100(self, storage):
        for i in range(110):
            storage.upsert_path_record(CONTACT_HEX, f"path{i:03d}", 1, 1, success=True)
        history = storage.get_path_history(CONTACT_HEX)
        assert len(history) == 100


# ─── Migration / schema ───────────────────────────────────────


class TestMigration:
    def test_tables_created(self, storage):
        tables = {
            row[0]
            for row in storage._db.execute(
                "SELECT name FROM sqlite_master WHERE type='table'"
            ).fetchall()
        }
        for expected in (
            "contacts",
            "messages",
            "channels",
            "channel_messages",
            "path_history",
            "room_passwords",
            "settings",
            "regions",
            "blocked_channel_senders",
        ):
            assert expected in tables, f"Missing table: {expected}"

    def test_contacts_has_migrated_columns(self, storage):
        cols = {row[1] for row in storage._db.execute("PRAGMA table_info(contacts)").fetchall()}
        for col in (
            "last_read_at",
            "path_hash_size",
            "device_lastmod",
            "force_flood",
            "auto_reset_path",
            "smart_route_selection",
        ):
            assert col in cols, f"Missing column: {col}"


# ─── Blocked Channel Senders ─────────────────────────────────


class TestBlockedChannelSenders:
    def test_empty_by_default(self, storage):
        assert storage.get_blocked_channel_senders() == set()

    def test_block_sender(self, storage):
        storage.block_channel_sender("Spammer")
        assert storage.is_channel_sender_blocked("Spammer") is True
        assert "Spammer" in storage.get_blocked_channel_senders()

    def test_unblock_sender(self, storage):
        storage.block_channel_sender("Spammer")
        storage.unblock_channel_sender("Spammer")
        assert storage.is_channel_sender_blocked("Spammer") is False
        assert storage.get_blocked_channel_senders() == set()

    def test_block_duplicate_is_noop(self, storage):
        storage.block_channel_sender("Spammer")
        storage.block_channel_sender("Spammer")
        assert storage.get_blocked_channel_senders() == {"Spammer"}

    def test_unblock_nonexistent_is_noop(self, storage):
        storage.unblock_channel_sender("Nobody")
        assert storage.get_blocked_channel_senders() == set()

    def test_multiple_blocked(self, storage):
        storage.block_channel_sender("Alice")
        storage.block_channel_sender("Bob")
        blocked = storage.get_blocked_channel_senders()
        assert blocked == {"Alice", "Bob"}

    def test_get_channel_senders(self, storage):
        storage.save_channel(Channel(index=0, name="General", psk=bytes(16)))
        storage._db.execute(
            "INSERT INTO channel_messages (channel_index, sender_name, text, timestamp, is_outgoing) "
            "VALUES (0, 'Alice', 'hi', 100.0, 0)",
        )
        storage._db.execute(
            "INSERT INTO channel_messages (channel_index, sender_name, text, timestamp, is_outgoing) "
            "VALUES (0, 'Alice', 'hello', 200.0, 0)",
        )
        storage._db.execute(
            "INSERT INTO channel_messages (channel_index, sender_name, text, timestamp, is_outgoing) "
            "VALUES (0, 'Bob', 'hey', 150.0, 0)",
        )
        storage._db.commit()
        senders = storage.get_channel_senders(0)
        assert len(senders) == 2
        assert senders[0]["sender_name"] == "Alice"
        assert senders[0]["message_count"] == 2
        assert senders[0]["last_seen"] == 200.0
        assert senders[1]["sender_name"] == "Bob"
        assert senders[1]["message_count"] == 1

    def test_get_channel_senders_excludes_outgoing(self, storage):
        storage.save_channel(Channel(index=0, name="General", psk=bytes(16)))
        storage._db.execute(
            "INSERT INTO channel_messages (channel_index, sender_name, text, timestamp, is_outgoing) "
            "VALUES (0, 'Me', 'hi', 100.0, 1)",
        )
        storage._db.commit()
        senders = storage.get_channel_senders(0)
        assert len(senders) == 0

    def test_get_channel_participants_includes_blocked_without_messages(self, storage):
        storage.save_channel(Channel(index=0, name="General", psk=bytes(16)))
        storage._db.execute(
            "INSERT INTO channel_messages (channel_index, sender_name, text, timestamp, is_outgoing) "
            "VALUES (0, 'Alice', 'hi', 100.0, 0)",
        )
        storage._db.commit()
        storage.block_channel_sender("Alice")
        storage.block_channel_sender("Ghost")
        participants = storage.get_channel_participants(0)
        assert len(participants) == 2
        alice = next(p for p in participants if p["sender_name"] == "Alice")
        assert alice["message_count"] == 1
        assert alice["is_blocked"] is True
        ghost = next(p for p in participants if p["sender_name"] == "Ghost")
        assert ghost["message_count"] == 0
        assert ghost["last_seen"] is None
        assert ghost["is_blocked"] is True

    def test_get_channel_participants_unblocked_not_flagged(self, storage):
        storage.save_channel(Channel(index=0, name="General", psk=bytes(16)))
        storage._db.execute(
            "INSERT INTO channel_messages (channel_index, sender_name, text, timestamp, is_outgoing) "
            "VALUES (0, 'Bob', 'hey', 100.0, 0)",
        )
        storage._db.commit()
        participants = storage.get_channel_participants(0)
        assert len(participants) == 1
        assert participants[0]["is_blocked"] is False

    def test_delete_messages_by_sender(self, storage):
        storage.save_channel(Channel(index=0, name="General", psk=bytes(16)))
        storage._db.execute(
            "INSERT INTO channel_messages (channel_index, sender_name, text, timestamp, is_outgoing) "
            "VALUES (0, 'Spammer', 'spam1', 100.0, 0)",
        )
        storage._db.execute(
            "INSERT INTO channel_messages (channel_index, sender_name, text, timestamp, is_outgoing) "
            "VALUES (0, 'Spammer', 'spam2', 200.0, 0)",
        )
        storage._db.execute(
            "INSERT INTO channel_messages (channel_index, sender_name, text, timestamp, is_outgoing) "
            "VALUES (0, 'Alice', 'hello', 150.0, 0)",
        )
        storage._db.commit()
        storage.delete_channel_messages_by_sender(0, "Spammer")
        msgs = storage.get_channel_messages(0)
        assert len(msgs) == 1
        assert msgs[0].sender_name == "Alice"


# ─── Channel Retention ──────────────────────────────────────


class TestChannelRetention:
    def test_default_retention_days(self, storage):
        storage.save_channel(Channel(index=0, name="General", psk=bytes(16)))
        ch = storage.get_channel(0)
        assert ch.retention_days == 0

    def test_set_channel_retention_days(self, storage):
        storage.save_channel(Channel(index=0, name="General", psk=bytes(16)))
        storage.set_channel_retention_days(0, 30)
        ch = storage.get_channel(0)
        assert ch.retention_days == 30

    def test_cleanup_prunes_old_messages(self, storage):
        import time as _time

        now = _time.time()
        storage.save_channel(Channel(index=0, name="Test", psk=bytes(16), retention_days=7))
        storage._db.execute(
            "INSERT INTO channel_messages (channel_index, sender_name, text, timestamp, is_outgoing) "
            "VALUES (0, 'Alice', 'old', ?, 0)",
            (now - 8 * 86400,),
        )
        storage._db.execute(
            "INSERT INTO channel_messages (channel_index, sender_name, text, timestamp, is_outgoing) "
            "VALUES (0, 'Bob', 'recent', ?, 0)",
            (now - 3 * 86400,),
        )
        storage._db.commit()
        storage._cleanup_old_data()
        msgs = storage.get_channel_messages(0)
        assert len(msgs) == 1
        assert msgs[0].sender_name == "Bob"

    def test_cleanup_skips_forever_channels(self, storage):
        import time as _time

        now = _time.time()
        storage.save_channel(Channel(index=0, name="Permanent", psk=bytes(16), retention_days=0))
        storage._db.execute(
            "INSERT INTO channel_messages (channel_index, sender_name, text, timestamp, is_outgoing) "
            "VALUES (0, 'Alice', 'ancient', ?, 0)",
            (now - 365 * 86400,),
        )
        storage._db.commit()
        storage._cleanup_old_data()
        msgs = storage.get_channel_messages(0)
        assert len(msgs) == 1

    def test_clear_channel_messages(self, storage):
        storage.save_channel(Channel(index=0, name="General", psk=bytes(16)))
        storage.save_channel(Channel(index=1, name="Other", psk=bytes(16)))
        storage._db.execute(
            "INSERT INTO channel_messages (channel_index, sender_name, text, timestamp, is_outgoing) "
            "VALUES (0, 'Alice', 'hi', 100.0, 0)",
        )
        storage._db.execute(
            "INSERT INTO channel_messages (channel_index, sender_name, text, timestamp, is_outgoing) "
            "VALUES (0, 'Bob', 'hey', 200.0, 0)",
        )
        storage._db.execute(
            "INSERT INTO channel_messages (channel_index, sender_name, text, timestamp, is_outgoing) "
            "VALUES (1, 'Carol', 'hello', 150.0, 0)",
        )
        storage._db.commit()
        storage.clear_channel_messages(0)
        assert len(storage.get_channel_messages(0)) == 0
        assert len(storage.get_channel_messages(1)) == 1


# ─── find_contact_by_prefix ──────────────────────────────────


class TestFindContactByPrefix:
    def test_finds_matching_contact(self, storage):
        key = bytes.fromhex(CONTACT_HEX)
        storage.save_contact(Contact(public_key=key, name="Alice"))
        found = storage.find_contact_by_prefix(key[:6])
        assert found is not None
        assert found.public_key_hex == CONTACT_HEX

    def test_no_match_returns_none(self, storage):
        storage.save_contact(Contact(public_key=bytes.fromhex(CONTACT_HEX), name="Alice"))
        found = storage.find_contact_by_prefix(bytes.fromhex("bb" * 6))
        assert found is None

    def test_prefix_ending_with_f(self, storage):
        key_hex = "ff" * 32
        storage.save_contact(Contact(public_key=bytes.fromhex(key_hex), name="Max"))
        found = storage.find_contact_by_prefix(bytes.fromhex(key_hex[:12]))
        assert found is not None
        assert found.name == "Max"

    def test_multiple_contacts_returns_one(self, storage):
        storage.save_contact(Contact(public_key=bytes.fromhex(CONTACT_HEX), name="Alice"))
        storage.save_contact(Contact(public_key=bytes.fromhex(OTHER_HEX), name="Bob"))
        found = storage.find_contact_by_prefix(bytes.fromhex(CONTACT_HEX[:12]))
        assert found is not None
        assert found.public_key_hex == CONTACT_HEX


# ─── fail_orphaned_messages ──────────────────────────────────


def _insert_msg_with_status(storage, contact_hex, text, ts, status, is_outgoing=True):
    storage._db.execute(
        """INSERT INTO messages
        (contact_key_hex, sender_key_hex, text, timestamp, is_outgoing, is_cli, status, message_id, room_sender_name)
        VALUES (?, ?, ?, ?, ?, 0, ?, NULL, '')""",
        (contact_hex, contact_hex, text, ts, int(is_outgoing), int(status)),
    )
    storage._db.commit()


class TestFailOrphanedMessages:
    def test_pending_marked_as_failed(self, storage):
        _insert_msg_with_status(storage, CONTACT_HEX, "msg1", 100.0, MessageStatus.PENDING)
        storage.fail_orphaned_messages()
        row = storage._db.execute("SELECT status FROM messages").fetchone()
        assert row[0] == int(MessageStatus.FAILED)

    def test_sent_marked_as_failed(self, storage):
        _insert_msg_with_status(storage, CONTACT_HEX, "msg1", 100.0, MessageStatus.SENT)
        storage.fail_orphaned_messages()
        row = storage._db.execute("SELECT status FROM messages").fetchone()
        assert row[0] == int(MessageStatus.FAILED)

    def test_delivered_not_affected(self, storage):
        _insert_msg_with_status(storage, CONTACT_HEX, "msg1", 100.0, MessageStatus.DELIVERED)
        storage.fail_orphaned_messages()
        row = storage._db.execute("SELECT status FROM messages").fetchone()
        assert row[0] == int(MessageStatus.DELIVERED)

    def test_incoming_not_affected(self, storage):
        _insert_msg_with_status(
            storage,
            CONTACT_HEX,
            "msg1",
            100.0,
            MessageStatus.PENDING,
            is_outgoing=False,
        )
        storage.fail_orphaned_messages()
        row = storage._db.execute("SELECT status FROM messages").fetchone()
        assert row[0] == int(MessageStatus.PENDING)

    def test_mixed_messages(self, storage):
        _insert_msg_with_status(storage, CONTACT_HEX, "a", 100.0, MessageStatus.PENDING)
        _insert_msg_with_status(storage, CONTACT_HEX, "b", 200.0, MessageStatus.SENT)
        _insert_msg_with_status(storage, CONTACT_HEX, "c", 300.0, MessageStatus.DELIVERED)
        storage.fail_orphaned_messages()
        rows = storage._db.execute(
            "SELECT text, status FROM messages ORDER BY timestamp"
        ).fetchall()
        assert rows[0][1] == int(MessageStatus.FAILED)
        assert rows[1][1] == int(MessageStatus.FAILED)
        assert rows[2][1] == int(MessageStatus.DELIVERED)


# ─── Storage.batch() ─────────────────────────────────────────


class TestBatch:
    def test_defers_commit(self, storage):
        storage.save_contact(Contact(public_key=bytes.fromhex(CONTACT_HEX), name="Alice"))
        with storage.batch():
            storage._db.execute(
                "UPDATE contacts SET name = ? WHERE public_key_hex = ?",
                ("Updated", CONTACT_HEX),
            )
            storage._commit()
            assert storage._batch_depth == 1

        assert storage._batch_depth == 0

    def test_nested_batch(self, storage):
        with storage.batch():
            assert storage._batch_depth == 1
            with storage.batch():
                assert storage._batch_depth == 2
            assert storage._batch_depth == 1
        assert storage._batch_depth == 0

    def test_depth_restores_after_exception(self, storage):
        try:
            with storage.batch():
                assert storage._batch_depth == 1
                raise ValueError("test error")
        except ValueError:
            pass
        assert storage._batch_depth == 0

    def test_commit_inside_batch_is_noop(self, storage):
        storage.save_contact(Contact(public_key=bytes.fromhex(CONTACT_HEX), name="Alice"))
        with storage.batch():
            storage._db.execute(
                "UPDATE contacts SET name = ? WHERE public_key_hex = ?",
                ("BatchUpdated", CONTACT_HEX),
            )
            storage._commit()
            assert storage._batch_depth == 1



# Copyright 2026, Jiri Eischmann <jiri@eischmann.cz> and the meshy contributors
# SPDX-License-Identifier: GPL-3.0-or-later

"""Sensor telemetry view for sensor-type contacts."""

import logging

import gi

gi.require_version("Gtk", "4.0")
gi.require_version("Adw", "1")
from gi.repository import Adw, GLib, Gtk
from meshy.models import Contact, ContactType
from meshy.utils import format_number

log = logging.getLogger(__name__)

_REQUEST_TIMEOUT_MS = 15000
_AUTO_REFRESH_MS = 30000


@Gtk.Template(resource_path="/page/codeberg/sesivany/Meshy/ui/sensor-view.ui")
class SensorView(Gtk.Box):
    __gtype_name__ = "MeshySensorView"

    _placeholder = Gtk.Template.Child("placeholder")
    _scroll = Gtk.Template.Child("scroll")
    _content_box = Gtk.Template.Child("content_box")
    _status_group = Gtk.Template.Child("status_group")
    _status_row = Gtk.Template.Child("status_row")
    _map_frame = Gtk.Template.Child("map_frame")
    _action_bar = Gtk.Template.Child("action_bar")
    _spinner = Gtk.Template.Child("spinner")
    _refresh_btn = Gtk.Template.Child("refresh_btn")
    _login_btn = Gtk.Template.Child("login_btn")

    def __init__(self, window, **kwargs):
        super().__init__(**kwargs)
        self._window = window
        self._contact: Contact | None = None
        self._timeout_id: int | None = None
        self._auto_refresh_id: int | None = None
        self._map_widget = None
        self._dynamic_groups: list[Gtk.Widget] = []

        self._refresh_btn.connect("clicked", lambda *_: self._request_telemetry())
        self._login_btn.connect("clicked", self._on_login_clicked)

    def set_sensor(self, contact: Contact):
        self.clear()
        self._contact = contact
        self._placeholder.set_visible(False)
        self._scroll.set_visible(True)
        self._action_bar.set_visible(True)
        self._status_row.set_subtitle(_("Requesting..."))
        self._status_group.set_visible(True)
        self._login_btn.set_visible(True)
        self._request_telemetry()
        self._auto_refresh_id = GLib.timeout_add(_AUTO_REFRESH_MS, self._on_auto_refresh)

    def clear(self):
        self._cancel_timeout()
        self._cancel_auto_refresh()
        self._cleanup_map()
        self._remove_dynamic_groups()
        self._contact = None
        self._status_row.set_subtitle("--")
        self._status_group.set_visible(True)
        self._placeholder.set_visible(True)
        self._scroll.set_visible(False)
        self._action_bar.set_visible(False)
        self._spinner.set_visible(False)
        self._spinner.set_spinning(False)
        self._login_btn.set_visible(False)

    def _request_telemetry(self):
        if not self._contact:
            return
        self._cancel_timeout()
        self._spinner.set_visible(True)
        self._spinner.set_spinning(True)
        self._refresh_btn.set_sensitive(False)
        self._timeout_id = GLib.timeout_add(_REQUEST_TIMEOUT_MS, self._on_timeout)
        self._window.get_telemetry(
            self._contact.public_key,
            self._on_telemetry_response,
            contact_type=ContactType.SENSOR,
        )

    def _on_telemetry_response(self, lpp_data):
        self._cancel_timeout()
        self._spinner.set_visible(False)
        self._spinner.set_spinning(False)
        self._refresh_btn.set_sensitive(True)

        if not lpp_data:
            self._status_row.set_subtitle(_("No telemetry data"))
            return

        from meshy.protocol import parse_cayenne_lpp

        entries = parse_cayenne_lpp(lpp_data)
        if not entries:
            self._status_row.set_subtitle(_("No telemetry data"))
            return

        self._display_telemetry(entries)

    def _display_telemetry(self, entries: list[dict]):
        from meshy.models import battery_percent_from_mv
        from meshy.protocol import LPP_GPS, LPP_VOLTAGE

        self._remove_dynamic_groups()
        self._cleanup_map()
        self._status_group.set_visible(False)

        channels: dict[int, list[dict]] = {}
        for entry in entries:
            channels.setdefault(entry["channel"], []).append(entry)

        gps_entry = None
        for ch in sorted(channels):
            group = Adw.PreferencesGroup(title=_("Channel {}").format(ch))

            for entry in channels[ch]:
                val = entry["value"]
                if isinstance(val, dict):
                    if entry["type"] == LPP_GPS:
                        gps_entry = entry
                        alt_str = (
                            f", alt {format_number(val['alt'], 0)} m"
                            if abs(val.get("alt", 0)) > 0.01
                            else ""
                        )
                        subtitle = f"{val['lat']:.4f}, {val['lon']:.4f}{alt_str}"
                    else:
                        subtitle = ", ".join(f"{k}: {v}" for k, v in val.items())
                elif entry["type"] == LPP_VOLTAGE and ch == 1:
                    pct = battery_percent_from_mv(int(val * 1000))
                    subtitle = f"{pct}% ({format_number(val, 2)} V)"
                elif isinstance(val, float):
                    subtitle = f"{format_number(val, 2)} {entry['unit']}"
                else:
                    subtitle = f"{val} {entry['unit']}"

                title = _("Battery") if entry["type"] == LPP_VOLTAGE and ch == 1 else entry["name"]
                row = Adw.ActionRow(title=title, subtitle=subtitle)
                group.add(row)

            insert_pos = len(self._dynamic_groups) + 1
            self._content_box.insert_child_after(
                group, self._status_group if insert_pos == 1 else self._dynamic_groups[-1]
            )
            self._dynamic_groups.append(group)

        if gps_entry:
            self._show_map(gps_entry)

    def _show_map(self, gps_entry: dict):
        try:
            from meshy.views import create_shumate_map
            from meshy.views.map_view import _MARKER_COLORS, _MARKER_ICONS, _SELF_COLOR, MapView

            lat = gps_entry["value"]["lat"]
            lon = gps_entry["value"]["lon"]

            map_widget, marker_layer, viewport = create_shumate_map(zoom_level=14, lat=lat, lon=lon)
            map_widget.set_vexpand(False)
            map_widget.set_size_request(-1, 250)

            r, g, b = _MARKER_COLORS.get(self._contact.type, (0.5, 0.5, 0.5))
            icon_name = _MARKER_ICONS.get(self._contact.type, "avatar-default-symbolic")
            marker = MapView.create_marker(r, g, b, icon_name, self._contact.name, lat, lon)
            marker_layer.add_marker(marker)

            lats, lons = [lat], [lon]
            si = self._window.self_info
            self_lat = si.get("adv_lat", 0.0)
            self_lon = si.get("adv_lon", 0.0)
            if abs(self_lat) > 1e-6 or abs(self_lon) > 1e-6:
                self_name = si.get("name", "") or _("My Device")
                self_marker = MapView.create_marker(
                    *_SELF_COLOR, "avatar-default-symbolic", self_name, self_lat, self_lon
                )
                marker_layer.add_marker(self_marker)
                lats.append(self_lat)
                lons.append(self_lon)

            MapView.fit_viewport(viewport, lats, lons)
            self._map_frame.set_child(map_widget)
            self._map_frame.set_visible(True)
            self._map_widget = map_widget
        except Exception as e:
            log.error("Failed to create map for sensor telemetry: %s", e, exc_info=True)
            lat = gps_entry["value"]["lat"]
            lon = gps_entry["value"]["lon"]
            from meshy.platforms import is_macos

            label = Gtk.Label()
            if is_macos():
                label.set_markup(f"<b>GPS Coordinates:</b>\nLat: {lat:.6f}, Lon: {lon:.6f}")
            else:
                label.set_markup(f"<b>GPS Coordinates:</b>\n📍 {lat:.6f}, {lon:.6f}")
            self._map_frame.set_child(label)
            self._map_frame.set_visible(True)

    def _on_login_clicked(self, btn):
        if not self._contact:
            return

        builder = Gtk.Builder.new_from_resource(
            "/page/codeberg/sesivany/Meshy/ui/sensor-login-dialog.ui"
        )
        dialog = builder.get_object("dialog")
        pwd_entry = builder.get_object("pwd_entry")
        login_btn = builder.get_object("login_btn")
        status_label = builder.get_object("status_label")

        dialog.set_title(_("Login to {}").format(self._contact.name))

        contact = self._contact

        def _on_login(*_args):
            password = pwd_entry.get_text().strip()
            if not password:
                return
            login_btn.set_sensitive(False)
            status_label.set_label(_("Logging in..."))

            def _on_login_result(success, login_timestamp, is_admin, acl_perms=3):
                if success:
                    dialog.close()
                    self._window.mark_room_logged_in(contact.public_key_hex, acl_perms)
                    self._window.show_toast(_("Logged in to {}").format(contact.name))
                else:
                    login_btn.set_sensitive(True)
                    status_label.set_label(_("Login failed"))

            self._window.send_login(contact.public_key, password, _on_login_result)

        login_btn.connect("clicked", _on_login)
        pwd_entry.connect("apply", _on_login)
        dialog.present(self._window)

    def _on_timeout(self):
        self._timeout_id = None
        self._spinner.set_visible(False)
        self._spinner.set_spinning(False)
        self._refresh_btn.set_sensitive(True)
        if not self._dynamic_groups:
            self._status_row.set_subtitle(_("No response (timed out)"))
        return False

    def _on_auto_refresh(self):
        if self._contact:
            self._request_telemetry()
            return True
        return False

    def _cancel_timeout(self):
        if self._timeout_id is not None:
            GLib.source_remove(self._timeout_id)
            self._timeout_id = None

    def _cancel_auto_refresh(self):
        if self._auto_refresh_id is not None:
            GLib.source_remove(self._auto_refresh_id)
            self._auto_refresh_id = None

    def _cleanup_map(self):
        if self._map_widget:
            from meshy.views import deregister_map_widget

            deregister_map_widget(self._map_widget)
            self._map_widget = None
        self._map_frame.set_child(None)
        self._map_frame.set_visible(False)

    def _remove_dynamic_groups(self):
        for group in self._dynamic_groups:
            self._content_box.remove(group)
        self._dynamic_groups.clear()



# Copyright 2026, Jiri Eischmann <jiri@eischmann.cz> and the meshy contributors
# SPDX-License-Identifier: GPL-3.0-or-later

import gettext
import locale
import math
import re
import unicodedata


def ngettext(singular, plural, n):
    return gettext.dngettext("meshy", singular, plural, n)


from gi.repository import GLib

_URL_RE = re.compile(r'(https?://[^\s<>"\')\]]+)')
_SHARED_CONTACT_RE = re.compile(r"^<([0-9a-fA-F]{64}):(\d+):(.+)>$")
_SHARED_LOCATION_RE = re.compile(r"^(-?\d+\.\d+),\s*(-?\d+\.\d+)$")


def linkify(text: str) -> str:
    escaped = GLib.markup_escape_text(text)

    def _replace(m):
        url = m.group(1)
        trailing = ""
        while url and url[-1] in ".,;:!?":
            trailing = url[-1] + trailing
            url = url[:-1]
        return f'<a href="{url}">{url}</a>{trailing}'

    return _URL_RE.sub(_replace, escaped)


def parse_shared_contact(text: str) -> tuple[str, int, str] | None:
    m = _SHARED_CONTACT_RE.match(text.strip())
    if not m:
        return None
    return (m.group(1), int(m.group(2)), m.group(3))


def parse_shared_location(text: str) -> tuple[float, float] | None:
    m = _SHARED_LOCATION_RE.match(text.strip())
    if not m:
        return None
    lat, lon = float(m.group(1)), float(m.group(2))
    if -90 <= lat <= 90 and -180 <= lon <= 180:
        return (lat, lon)
    return None


def haversine_distance(lat1, lon1, lat2, lon2) -> float:
    """Return the great-circle distance in km between two points."""
    R = 6371.0
    dlat = math.radians(lat2 - lat1)
    dlon = math.radians(lon2 - lon1)
    a = (
        math.sin(dlat / 2) ** 2
        + math.cos(math.radians(lat1)) * math.cos(math.radians(lat2)) * math.sin(dlon / 2) ** 2
    )
    return R * 2 * math.atan2(math.sqrt(a), math.sqrt(1 - a))


def format_number(value: float, decimals: int = 1, sign: bool = False) -> str:
    fmt = f"%{'+' if sign else ''}.{decimals}f"
    return locale.format_string(fmt, value)


def parse_number(text: str) -> float:
    try:
        return locale.atof(text)
    except ValueError:
        return float(text)


def format_trip_time(ms: int | None) -> str:
    if not ms:
        return ""
    return locale.format_string("%.1f", ms / 1000) + " " + "s"


def format_duration(secs):
    d, secs = divmod(int(secs), 86400)
    h, secs = divmod(secs, 3600)
    m, secs = divmod(secs, 60)
    parts = []
    if d:
        parts.append(_("{d}d").format(d=d))
    if h:
        parts.append(_("{h}h").format(h=h))
    if m:
        parts.append(_("{m}m").format(m=m))
    parts.append(_("{secs}s").format(secs=secs))
    return " ".join(parts)


def format_time_ago(secs):
    secs = int(secs)
    if secs < 60:
        return ngettext("{} sec ago", "{} secs ago", secs).format(secs)
    if secs < 3600:
        mins = secs // 60
        return ngettext("{} min ago", "{} min ago", mins).format(mins)
    if secs < 86400:
        hours = secs // 3600
        return ngettext("{} hour ago", "{} hours ago", hours).format(hours)
    days = secs // 86400
    return ngettext("{} day ago", "{} days ago", days).format(days)


class TimeoutManager:
    def __init__(self):
        self._timers: dict[str, int] = {}

    def start(self, name, ms, callback, *args):
        self.cancel(name)
        self._timers[name] = GLib.timeout_add(ms, self._fire, name, callback, args)

    def _fire(self, name, callback, args):
        self._timers.pop(name, None)
        callback(*args)
        return False

    def cancel(self, name):
        tid = self._timers.pop(name, None)
        if tid is not None:
            GLib.source_remove(tid)

    def is_active(self, name):
        return name in self._timers

    def cancel_all(self):
        for tid in self._timers.values():
            GLib.source_remove(tid)
        self._timers.clear()


def clear_children(container):
    while child := container.get_first_child():
        container.remove(child)


def extract_emoji(name: str) -> str | None:
    if not name:
        return None
    for char in name:
        cat = unicodedata.category(char)
        if cat == "So":
            return char
        cp = ord(char)
        if 0x1F600 <= cp <= 0x1F64F:
            return char
        if 0x1F300 <= cp <= 0x1F5FF:
            return char
        if 0x1F680 <= cp <= 0x1F6FF:
            return char
        if 0x1F900 <= cp <= 0x1F9FF:
            return char
        if 0x1FA00 <= cp <= 0x1FAFF:
            return char
        if 0x2600 <= cp <= 0x27BF:
            return char
        if 0xFE00 <= cp <= 0xFE0F:
            continue
        if cp == 0x200D:
            continue
    return None



"""Tests for smart routing: standalone CONTACT handling, PATH_UPDATED parsing,
and flood attribution chain."""

import struct
from unittest.mock import MagicMock, patch

from meshy.frame_handler import FrameHandler
from meshy.message_controller import MessageController
from meshy.protocol import (
    PUSH_CODE_PATH_UPDATED,
    RESP_CODE_CONTACT,
)
from meshy.storage import Storage

PUB_KEY = bytes(range(32))
PUB_KEY_HEX = PUB_KEY.hex()


def _contact_dict(path_len=2, path_hex="aabb", lastmod=100):
    return {
        "public_key": PUB_KEY_HEX,
        "type": 1,
        "flags": 0,
        "out_path_len": path_len,
        "path_hash_size": 1,
        "out_path": path_hex,
        "adv_name": "Test",
        "adv_lat": 0.0,
        "adv_lon": 0.0,
        "last_advert": 0,
        "lastmod": lastmod,
    }


def _build_contact_frame(pub_key=PUB_KEY, path_len_byte=2, path=None):
    """Build a raw CONTACT frame (code 3) matching the firmware format."""
    frame = bytes([RESP_CODE_CONTACT])
    frame += pub_key
    frame += bytes([1])  # type = CHAT
    frame += bytes([0])  # flags
    frame += bytes([path_len_byte])
    frame += (path or bytes(64)).ljust(64, b"\x00")[:64]
    frame += b"TestNode" + bytes(24)  # name (32 bytes)
    frame += struct.pack("<I", 1700000000)  # last_advert
    frame += struct.pack("<i", 50000000)  # lat
    frame += struct.pack("<i", 14000000)  # lon
    frame += struct.pack("<I", 100)  # lastmod
    return frame


def _make_win(storage):
    win = MagicMock()
    win._storage = storage
    win._contacts = []
    win._contacts_by_key = {}
    win._conn_ctrl.contacts_iter_busy = False
    win._conn_ctrl.synced_contact_keys = None
    win._conn_ctrl.sync_contact_current = 0
    win._conn_ctrl.syncing = False
    win._device_info = None
    win._login_callback = None
    win._ignore_next_err = False
    win._msg_ctrl = MessageController(win)
    win._region_callbacks = None
    return win


# ─── Standalone CONTACT frame (getContactByKey response) ─────


class TestStandaloneContactFrame:
    def setup_method(self):
        self.storage = Storage(db_path=":memory:")
        self.win = _make_win(self.storage)
        self.fh = FrameHandler(self.win)

    def teardown_method(self):
        self.storage.close()

    def test_standalone_contact_updates_contact(self):
        """A CONTACT frame outside bulk sync should call add_or_update_contact."""
        self.win._conn_ctrl.contacts_iter_busy = False
        frame = _build_contact_frame()
        with patch.object(self.fh, "add_or_update_contact") as mock_add:
            self.fh.handle_frame(frame)
            mock_add.assert_called_once()

    def test_standalone_contact_skips_sync_bookkeeping(self):
        """With contacts_iter_busy=False, sync counter must not increment."""
        self.win._conn_ctrl.contacts_iter_busy = False
        self.win._conn_ctrl.sync_contact_current = 0
        frame = _build_contact_frame()
        with patch.object(self.fh, "add_or_update_contact"):
            self.fh.handle_frame(frame)
        assert self.win._conn_ctrl.sync_contact_current == 0
        self.win._conn_ctrl._start_contacts_sync_timeout.assert_not_called()

    def test_bulk_sync_contact_does_bookkeeping(self):
        """With contacts_iter_busy=True, sync counter must increment."""
        self.win._conn_ctrl.contacts_iter_busy = True
        self.win._conn_ctrl.sync_contact_current = 0
        self.win._conn_ctrl.synced_contact_keys = set()
        frame = _build_contact_frame()
        with (
            patch.object(self.fh, "add_or_update_contact"),
            patch.object(self.fh, "dict_to_contact") as mock_dtc,
        ):
            mock_contact = MagicMock()
            mock_contact.public_key_hex = PUB_KEY_HEX
            mock_dtc.return_value = mock_contact
            self.fh.handle_frame(frame)
        assert self.win._conn_ctrl.sync_contact_current == 1
        self.win._conn_ctrl._start_contacts_sync_timeout.assert_called_once()
        assert PUB_KEY_HEX in self.win._conn_ctrl.synced_contact_keys


# ─── PATH_UPDATED parsing ────────────────────────────────────


class TestPathUpdatedParsing:
    def setup_method(self):
        self.storage = Storage(db_path=":memory:")
        self.win = _make_win(self.storage)
        self.fh = FrameHandler(self.win)

    def teardown_method(self):
        self.storage.close()

    def test_extracts_pub_key_and_fetches(self):
        """PATH_UPDATED with 32-byte pub key should call refresh_contact."""
        frame = bytes([PUSH_CODE_PATH_UPDATED]) + PUB_KEY
        with patch.object(self.fh, "refresh_contact") as mock_refresh:
            self.fh.handle_frame(frame)
        mock_refresh.assert_called_once_with(PUB_KEY_HEX)

    def test_short_frame_skips_fetch(self):
        """PATH_UPDATED with too-short data should fall back to bulk refetch."""
        frame = bytes([PUSH_CODE_PATH_UPDATED]) + bytes(10)
        with patch.object(self.fh, "refresh_contact") as mock_refresh:
            with patch.object(self.fh, "schedule_contacts_refetch") as mock_refetch:
                self.fh.handle_frame(frame)
            mock_refresh.assert_not_called()
            mock_refetch.assert_called_once()

    def test_still_triggers_bulk_refetch(self):
        """PATH_UPDATED with pub key calls refresh_contact which includes bulk refetch."""
        frame = bytes([PUSH_CODE_PATH_UPDATED]) + PUB_KEY
        with patch.object(self.fh, "send_get_contact_by_key"):
            with patch.object(self.fh, "schedule_contacts_refetch") as mock_refetch:
                self.fh.handle_frame(frame)
            mock_refetch.assert_called_once()


# ─── Flood attribution chain ─────────────────────────────────


class TestFloodAttributionChain:
    def setup_method(self):
        self.storage = Storage(db_path=":memory:")
        self.win = _make_win(self.storage)
        self.fh = FrameHandler(self.win)

    def teardown_method(self):
        self.storage.close()

    def test_attribution_consumed_when_contact_has_path(self):
        """pending_flood_attributions should be consumed and path recorded
        when add_or_update_contact sees path_length >= 0."""
        self.win._msg_ctrl.pending_flood_attributions[PUB_KEY_HEX] = [(5000, False)]
        self.fh.add_or_update_contact(_contact_dict(path_len=2, path_hex="aabb"))

        assert PUB_KEY_HEX not in self.win._msg_ctrl.pending_flood_attributions
        history = self.storage.get_path_history(PUB_KEY_HEX)
        assert len(history) == 1
        assert history[0]["last_trip_ms"] == 5000

    def test_attribution_not_consumed_when_flood(self):
        """pending_flood_attributions should remain when contact has path_length < 0."""
        self.win._msg_ctrl.pending_flood_attributions[PUB_KEY_HEX] = [(5000, False)]
        self.fh.add_or_update_contact(_contact_dict(path_len=-1, path_hex=""))

        assert PUB_KEY_HEX in self.win._msg_ctrl.pending_flood_attributions
        assert self.storage.get_path_history(PUB_KEY_HEX) == []

    def test_seed_from_contact_when_no_pending_attribution(self):
        """Without pending attributions, a contact with a path should be seeded."""
        self.fh.add_or_update_contact(_contact_dict(path_len=2, path_hex="ccdd"))

        history = self.storage.get_path_history(PUB_KEY_HEX)
        assert len(history) == 1
        assert history[0]["path_bytes"] == "ccdd"
        assert history[0]["route_weight"] == 1.0

    def test_late_attribution_uses_reduced_weight(self):
        """Late flood attribution should use GRACE_WEIGHT_INCREMENT (0.2),
        not the default 0.5, when updating an existing path record."""
        self.storage.upsert_path_record(
            PUB_KEY_HEX,
            "eeff",
            1,
            2,
            success=True,
            trip_ms=5000,
        )
        baseline = self.storage.get_path_history(PUB_KEY_HEX)[0]["route_weight"]

        self.win._msg_ctrl.pending_flood_attributions[PUB_KEY_HEX] = [(8000, True)]
        self.fh.add_or_update_contact(_contact_dict(path_len=2, path_hex="eeff"))

        history = self.storage.get_path_history(PUB_KEY_HEX)
        assert len(history) == 1
        assert history[0]["route_weight"] == baseline + 0.2
        assert history[0]["last_trip_ms"] == 8000



"""Tests for room server support: protocol builders, storage, frame handler, ACL parsing."""

import struct
import time
from datetime import datetime
from unittest.mock import MagicMock, patch

from meshy.frame_handler import FrameHandler
from meshy.models import Contact, ContactType, Message, MessageStatus
from meshy.protocol import (
    AUTO_ADD_CHAT,
    AUTO_ADD_ROOM_SERVER,
    CMD_LOGOUT,
    CMD_SEND_BINARY_REQ,
    CMD_SEND_LOGIN,
    CMD_SET_AUTO_ADD_CONFIG,
    PUSH_CODE_LOGIN_FAILED,
    PUSH_CODE_LOGIN_SUCCESS,
    REQ_TYPE_ACL,
    build_get_acl,
    build_logout,
    build_send_login,
    build_set_auto_add_config,
)
from meshy.repeater_helpers import (
    ACL_ROLES,
    ROLE_PERMS,
    parse_acl_response,
    parse_region_text,
)

_PUB_KEY = bytes(range(32))
_PREFIX = _PUB_KEY[:6]


def _make_contact(name="TestRoom", type_=ContactType.ROOM, key_hex=None):
    if key_hex is None:
        key_hex = "aa" * 32
    return Contact(public_key=bytes.fromhex(key_hex), name=name, type=type_)


def _make_win(storage):
    win = MagicMock()
    win._storage = storage
    win._self_info = {"name": "Me"}
    win._channels_view.on_path_variant_received.return_value = False
    win._channels = []
    win._channels_view_inst = None
    win._conn_ctrl.channel_sync_index = -1
    win._login_callback = None
    win._login_target_prefix = None
    win._login_pub_key = None
    win._login_was_flood_retry = False
    win._login_is_admin = False
    win._contacts = []
    win._pending_binary_requests = {}
    win._cli_pending_prefixes = {}
    win._cli_response_callback = None
    return win


def _build_login_success(prefix, perms=0x01, timestamp=1700000000, acl_perm=3, fw_ver=1):
    data = bytes([PUSH_CODE_LOGIN_SUCCESS, perms])
    data += prefix
    data += struct.pack("<I", timestamp)
    data += bytes([acl_perm, fw_ver])
    return data


def _build_login_failed(prefix):
    return bytes([PUSH_CODE_LOGIN_FAILED, 0x00]) + prefix


def _build_acl_response(entries):
    """Build ACL binary from list of (key_hex_12chars, perm) tuples."""
    data = bytes(6)  # code + reserved + tag
    for key_hex, perm in entries:
        data += bytes.fromhex(key_hex) + bytes([perm])
    return data


# ─── Protocol Builders ─────────────────────────────────────


class TestProtocolRoomBuilders:
    def test_build_send_login_layout(self):
        frame = build_send_login(_PUB_KEY, "secret")
        assert frame[0] == CMD_SEND_LOGIN
        assert frame[1:33] == _PUB_KEY
        assert frame[33:-1] == b"secret"
        assert frame[-1] == 0x00

    def test_build_send_login_length(self):
        pwd = "test123"
        frame = build_send_login(_PUB_KEY, pwd)
        assert len(frame) == 1 + 32 + len(pwd.encode()) + 1

    def test_build_send_login_unicode(self):
        frame = build_send_login(_PUB_KEY, "héslo")
        assert "héslo".encode() in frame

    def test_build_logout_layout(self):
        frame = build_logout(_PUB_KEY)
        assert frame[0] == CMD_LOGOUT
        assert frame[1:] == _PUB_KEY
        assert len(frame) == 33

    def test_build_get_acl_layout(self):
        frame = build_get_acl(_PUB_KEY)
        assert frame[0] == CMD_SEND_BINARY_REQ
        assert frame[1:33] == _PUB_KEY
        assert frame[33] == REQ_TYPE_ACL
        assert frame[34:36] == bytes([0x00, 0x00])
        assert len(frame) == 36

    def test_auto_add_room_only(self):
        frame = build_set_auto_add_config(room=True)
        assert frame[0] == CMD_SET_AUTO_ADD_CONFIG
        assert frame[1] == AUTO_ADD_ROOM_SERVER  # 0x08

    def test_auto_add_room_and_chat(self):
        frame = build_set_auto_add_config(room=True, chat=True)
        assert frame[1] == AUTO_ADD_ROOM_SERVER | AUTO_ADD_CHAT

    def test_auto_add_no_flags(self):
        frame = build_set_auto_add_config()
        assert frame[1] == 0x00

    def test_auto_add_max_hops_clamped(self):
        frame = build_set_auto_add_config(max_hops=100)
        assert frame[2] == 64


# ─── Storage: Room Passwords ──────────────────────────────


class TestStorageRoomPasswords:
    def test_save_and_get(self, storage):
        storage.save_room_password("aabbcc", "secret")
        assert storage.get_room_password("aabbcc") == "secret"

    def test_get_nonexistent(self, storage):
        assert storage.get_room_password("unknown") is None

    def test_update(self, storage):
        storage.save_room_password("aabbcc", "old")
        storage.save_room_password("aabbcc", "new")
        assert storage.get_room_password("aabbcc") == "new"

    def test_remove(self, storage):
        storage.save_room_password("aabbcc", "secret")
        storage.remove_room_password("aabbcc")
        assert storage.get_room_password("aabbcc") is None

    def test_remove_nonexistent(self, storage):
        storage.remove_room_password("nonexistent")

    def test_multiple_rooms(self, storage):
        storage.save_room_password("room1a", "pw1")
        storage.save_room_password("room2b", "pw2")
        assert storage.get_room_password("room1a") == "pw1"
        assert storage.get_room_password("room2b") == "pw2"


# ─── Storage: Room Sender Name ────────────────────────────


class TestStorageRoomSenderName:
    def test_room_sender_name_roundtrip(self, storage):
        contact = _make_contact()
        storage.save_contact(contact)
        msg = Message(
            sender_key=contact.public_key,
            text="hello",
            timestamp=datetime.now(),
            is_outgoing=False,
            status=MessageStatus.DELIVERED,
            room_sender_name="Alice",
        )
        storage.save_message(contact.public_key_hex, msg)
        messages = storage.get_messages(contact.public_key_hex)
        assert len(messages) == 1
        assert messages[0].room_sender_name == "Alice"

    def test_room_sender_name_default_empty(self, storage):
        contact = _make_contact()
        storage.save_contact(contact)
        msg = Message(
            sender_key=contact.public_key,
            text="hello",
            timestamp=datetime.now(),
            is_outgoing=False,
            status=MessageStatus.DELIVERED,
        )
        storage.save_message(contact.public_key_hex, msg)
        messages = storage.get_messages(contact.public_key_hex)
        assert messages[0].room_sender_name == ""


# ─── Frame Handler: Login Success ─────────────────────────


class TestFrameHandlerLoginSuccess:
    def setup_method(self, storage=None):
        self._storage_backup = storage

    @patch("meshy.frame_handler.GLib")
    def test_admin_login(self, mock_glib, storage):
        win = _make_win(storage)
        fh = FrameHandler(win)
        cb = MagicMock()
        win._login_callback = cb
        win._login_target_prefix = _PREFIX

        frame = _build_login_success(_PREFIX, perms=0x01, timestamp=1700000000, acl_perm=3)
        fh.handle_frame(frame)

        cb.assert_called_once_with(True, 1700000000, True, 3)

    @patch("meshy.frame_handler.GLib")
    def test_non_admin_login(self, mock_glib, storage):
        win = _make_win(storage)
        fh = FrameHandler(win)
        cb = MagicMock()
        win._login_callback = cb
        win._login_target_prefix = _PREFIX

        frame = _build_login_success(_PREFIX, perms=0x00, acl_perm=2)
        fh.handle_frame(frame)

        cb.assert_called_once()
        args = cb.call_args[0]
        assert args[0] is True
        assert args[2] is False  # is_admin
        assert args[3] == 2  # acl_permissions

    @patch("meshy.frame_handler.GLib")
    def test_short_frame_admin_defaults_to_3(self, mock_glib, storage):
        win = _make_win(storage)
        fh = FrameHandler(win)
        cb = MagicMock()
        win._login_callback = cb
        win._login_target_prefix = _PREFIX

        # 12 bytes: code + perms + prefix(6) + timestamp(4) — no ACL byte
        frame = bytes([PUSH_CODE_LOGIN_SUCCESS, 0x01]) + _PREFIX + struct.pack("<I", 100)
        fh.handle_frame(frame)

        args = cb.call_args[0]
        assert args[3] == 3  # admin defaults to 3

    @patch("meshy.frame_handler.GLib")
    def test_short_frame_non_admin_defaults_to_2(self, mock_glib, storage):
        win = _make_win(storage)
        fh = FrameHandler(win)
        cb = MagicMock()
        win._login_callback = cb
        win._login_target_prefix = _PREFIX

        frame = bytes([PUSH_CODE_LOGIN_SUCCESS, 0x00]) + _PREFIX + struct.pack("<I", 100)
        fh.handle_frame(frame)

        args = cb.call_args[0]
        assert args[3] == 2  # non-admin defaults to 2

    @patch("meshy.frame_handler.GLib")
    def test_prefix_mismatch(self, mock_glib, storage):
        win = _make_win(storage)
        fh = FrameHandler(win)
        cb = MagicMock()
        win._login_callback = cb
        win._login_target_prefix = b"\xff" * 6

        frame = _build_login_success(_PREFIX)
        fh.handle_frame(frame)

        cb.assert_not_called()

    @patch("meshy.frame_handler.GLib")
    def test_no_callback(self, mock_glib, storage):
        win = _make_win(storage)
        fh = FrameHandler(win)
        win._login_callback = None

        frame = _build_login_success(_PREFIX)
        fh.handle_frame(frame)  # should not raise

    @patch("meshy.frame_handler.GLib")
    def test_flood_retry_resets_path(self, mock_glib, storage):
        win = _make_win(storage)
        fh = FrameHandler(win)
        cb = MagicMock()
        win._login_callback = cb
        win._login_target_prefix = _PREFIX
        win._login_was_flood_retry = True
        win._login_pub_key = _PUB_KEY

        frame = _build_login_success(_PREFIX)
        fh.handle_frame(frame)

        cb.assert_called_once()
        win.send_frame.assert_called_once()

    @patch("meshy.frame_handler.GLib")
    def test_too_short_frame(self, mock_glib, storage):
        win = _make_win(storage)
        fh = FrameHandler(win)
        win._login_callback = MagicMock()
        win._login_target_prefix = _PREFIX

        frame = bytes([PUSH_CODE_LOGIN_SUCCESS, 0x01, 0x00, 0x00])
        result = fh.handle_frame(frame)
        assert result is False


# ─── Frame Handler: Login Failed ──────────────────────────


class TestFrameHandlerLoginFailed:
    @patch("meshy.frame_handler.GLib")
    def test_prefix_match(self, mock_glib, storage):
        win = _make_win(storage)
        fh = FrameHandler(win)
        cb = MagicMock()
        win._login_callback = cb
        win._login_target_prefix = _PREFIX

        frame = _build_login_failed(_PREFIX)
        fh.handle_frame(frame)

        cb.assert_called_once_with(False, None, False, 0)

    @patch("meshy.frame_handler.GLib")
    def test_prefix_mismatch(self, mock_glib, storage):
        win = _make_win(storage)
        fh = FrameHandler(win)
        cb = MagicMock()
        win._login_callback = cb
        win._login_target_prefix = b"\xff" * 6

        frame = _build_login_failed(_PREFIX)
        fh.handle_frame(frame)

        cb.assert_not_called()


# ─── Frame Handler: Room Sender Resolution ────────────────


class TestFrameHandlerRoomSender:
    @patch("meshy.frame_handler.GLib")
    def test_resolves_sender_name(self, mock_glib, storage):
        room_key = "bb" * 32
        sender_key = "cc" * 32
        room = _make_contact("MyRoom", ContactType.ROOM, room_key)
        sender = _make_contact("Alice", ContactType.CHAT, sender_key)
        storage.save_contact(room)
        storage.save_contact(sender)

        win = _make_win(storage)
        win._contacts = [room, sender]
        fh = FrameHandler(win)

        sig = sender_key[:8]  # 4-byte prefix as hex
        fh.handle_incoming_message(
            {
                "pubkey_prefix": room_key[:12],
                "txt_type": 2,
                "sender_timestamp": time.time(),
                "text": "hello",
                "signature": sig,
            }
        )

        saved_msg = storage.get_messages(room_key)
        assert len(saved_msg) == 1
        assert saved_msg[0].room_sender_name == "Alice"

    @patch("meshy.frame_handler.GLib")
    def test_unknown_sender_uses_signature(self, mock_glib, storage):
        room_key = "bb" * 32
        room = _make_contact("MyRoom", ContactType.ROOM, room_key)
        storage.save_contact(room)

        win = _make_win(storage)
        win._contacts = [room]
        fh = FrameHandler(win)

        fh.handle_incoming_message(
            {
                "pubkey_prefix": room_key[:12],
                "txt_type": 2,
                "sender_timestamp": time.time(),
                "text": "hello",
                "signature": "deadbeef",
            }
        )

        saved_msg = storage.get_messages(room_key)
        assert len(saved_msg) == 1
        assert saved_msg[0].room_sender_name == "deadbeef"

    @patch("meshy.frame_handler.GLib")
    def test_non_room_no_resolution(self, mock_glib, storage):
        chat_key = "dd" * 32
        contact = _make_contact("Bob", ContactType.CHAT, chat_key)
        storage.save_contact(contact)

        win = _make_win(storage)
        win._contacts = [contact]
        fh = FrameHandler(win)

        fh.handle_incoming_message(
            {
                "pubkey_prefix": chat_key[:12],
                "txt_type": 2,
                "sender_timestamp": time.time(),
                "text": "hi",
                "signature": "aabb",
            }
        )

        saved_msg = storage.get_messages(chat_key)
        assert len(saved_msg) == 1
        assert saved_msg[0].room_sender_name == ""

    @patch("meshy.frame_handler.GLib")
    def test_cli_message_no_resolution(self, mock_glib, storage):
        room_key = "bb" * 32
        sender_key = "cc" * 32
        room = _make_contact("MyRoom", ContactType.ROOM, room_key)
        sender = _make_contact("Alice", ContactType.CHAT, sender_key)
        storage.save_contact(room)

        win = _make_win(storage)
        win._contacts = [room, sender]
        win._cli_response_callback = None
        fh = FrameHandler(win)

        fh.handle_incoming_message(
            {
                "pubkey_prefix": room_key[:12],
                "txt_type": 1,  # CLI
                "sender_timestamp": time.time(),
                "text": "> OK",
                "signature": sender_key[:8],
            }
        )

        saved_msg = storage.get_messages(room_key)
        assert len(saved_msg) == 1
        assert saved_msg[0].room_sender_name == ""


# ─── ACL Response Parsing ─────────────────────────────────


class TestParseAclResponse:
    def test_single_entry(self):
        data = _build_acl_response([("aabbccddeeff", 3)])
        result = parse_acl_response(data)
        assert result == [{"key": "aabbccddeeff", "perm": 3}]

    def test_multiple_entries(self):
        data = _build_acl_response(
            [
                ("aabbccddeeff", 3),
                ("112233445566", 2),
                ("ffeeddccbbaa", 1),
            ]
        )
        result = parse_acl_response(data)
        assert len(result) == 3
        assert result[0]["key"] == "aabbccddeeff"
        assert result[1]["perm"] == 2
        assert result[2]["key"] == "ffeeddccbbaa"

    def test_zero_key_filtered(self):
        data = _build_acl_response(
            [
                ("aabbccddeeff", 3),
                ("000000000000", 1),
            ]
        )
        result = parse_acl_response(data)
        assert len(result) == 1

    def test_empty_payload(self):
        data = bytes(6)  # header only
        assert parse_acl_response(data) == []

    def test_partial_entry_ignored(self):
        data = bytes(6) + bytes(5)  # header + incomplete entry
        assert parse_acl_response(data) == []

    def test_permission_bitmask(self):
        data = _build_acl_response([("aabbccddeeff", 0x07)])
        result = parse_acl_response(data)
        assert ACL_ROLES[result[0]["perm"] & 3] == "Admin"


# ─── ACL Constants ─────────────────────────────────────────


class TestAclConstants:
    def test_acl_roles_complete(self):
        assert set(ACL_ROLES.keys()) == {1, 2, 3}

    def test_role_perms_mapping(self):
        assert ROLE_PERMS[0] == 3  # Admin
        assert ROLE_PERMS[1] == 2  # Read-Write
        assert ROLE_PERMS[2] == 1  # Read Only


# ─── Region Text Parsing ──────────────────────────────────


class TestParseRegionText:
    def test_simple_region(self):
        entries, _ = parse_region_text("US")
        assert entries == [{"name": "US", "indent": 0, "home": False, "flood": False}]

    def test_flood_suffix(self):
        entries, _ = parse_region_text("EU F")
        assert entries[0]["flood"] is True
        assert entries[0]["name"] == "EU"

    def test_home_suffix(self):
        entries, _ = parse_region_text("US^")
        assert entries[0]["home"] is True
        assert entries[0]["name"] == "US"

    def test_flood_and_home(self):
        entries, _ = parse_region_text("US^ F")
        assert entries[0]["flood"] is True
        assert entries[0]["home"] is True
        assert entries[0]["name"] == "US"

    def test_indented_regions(self):
        text = "*\n  US\n    East"
        entries, _ = parse_region_text(text)
        assert len(entries) == 3
        assert entries[0]["indent"] == 0
        assert entries[1]["indent"] == 2
        assert entries[2]["indent"] == 4

    def test_empty_lines_skipped(self):
        entries, _ = parse_region_text("US\n\nEU")
        assert len(entries) == 2

    def test_default_scope_parsed(self):
        _, default = parse_region_text("US", "blah blah default scope is US")
        assert default == "US"

    def test_default_scope_null(self):
        _, default = parse_region_text("US", "default scope is <null>")
        assert default == ""

    def test_default_scope_missing(self):
        _, default = parse_region_text("US", "")
        assert default == ""

    def test_empty_region_text(self):
        entries, default = parse_region_text("", "default scope is US")
        assert entries == []
        assert default == "US"



import struct

from meshy.models import Contact, ContactType, battery_percent_from_mv
from meshy.protocol import (
    CMD_SEND_BINARY_REQ,
    CMD_SEND_LOGIN,
    LPP_ACCELEROMETER,
    LPP_COLOUR,
    LPP_CONCENTRATION,
    LPP_CURRENT,
    LPP_DIGITAL_INPUT,
    LPP_DISTANCE,
    LPP_FREQUENCY,
    LPP_GPS,
    LPP_HUMIDITY,
    LPP_LUMINOSITY,
    LPP_PERCENTAGE,
    LPP_PRESSURE,
    LPP_SWITCH,
    LPP_TEMPERATURE,
    LPP_UNIX_TIME,
    LPP_VOLTAGE,
    REQ_TYPE_TELEMETRY,
    build_get_telemetry,
    build_send_login,
    parse_cayenne_lpp,
)

# ─── Contact model: sensor type ─────────────────────────────


class TestSensorContact:
    def _make(self, **kwargs):
        defaults = dict(public_key=bytes(32), name="Weather-1", type=ContactType.SENSOR)
        defaults.update(kwargs)
        return Contact(**defaults)

    def test_type_label(self):
        c = self._make()
        assert c.type_label == "Sensor"

    def test_type_value(self):
        assert ContactType.SENSOR == 4

    def test_public_key_hex(self):
        key = bytes(range(32))
        c = self._make(public_key=key)
        assert c.public_key_hex == key.hex()

    def test_path_label_flood(self):
        c = self._make(path_length=-1)
        assert c.path_label == "Flood"

    def test_path_label_direct(self):
        c = self._make(path_length=0)
        assert c.path_label == "Direct"

    def test_has_location_valid(self):
        c = self._make(latitude=50.0, longitude=14.0)
        assert c.has_location is True

    def test_has_location_none(self):
        c = self._make()
        assert c.has_location is False


# ─── Protocol: build_get_telemetry ───────────────────────────


class TestBuildGetTelemetry:
    def test_frame_layout(self):
        pub_key = bytes(range(32))
        frame = build_get_telemetry(pub_key)
        assert frame[0] == CMD_SEND_BINARY_REQ
        assert frame[1:33] == pub_key
        assert frame[33] == REQ_TYPE_TELEMETRY
        assert len(frame) == 34

    def test_different_keys(self):
        k1 = bytes([0xAA] * 32)
        k2 = bytes([0xBB] * 32)
        f1 = build_get_telemetry(k1)
        f2 = build_get_telemetry(k2)
        assert f1[1:33] != f2[1:33]
        assert f1[0] == f2[0] == CMD_SEND_BINARY_REQ
        assert f1[33] == f2[33] == REQ_TYPE_TELEMETRY


# ─── Protocol: build_send_login (sensor admin) ───────────────


class TestSensorLogin:
    def test_frame_layout(self):
        pub_key = bytes(range(32))
        frame = build_send_login(pub_key, "secret")
        assert frame[0] == CMD_SEND_LOGIN
        assert frame[1:33] == pub_key
        assert frame[33:-1] == b"secret"
        assert frame[-1] == 0  # null terminator

    def test_empty_password(self):
        frame = build_send_login(bytes(32), "")
        assert frame[-1] == 0
        assert len(frame) == 34  # cmd + 32 key + null


# ─── Cayenne LPP: sensor-typical payloads ────────────────────


class TestCayenneLPPSensorPayloads:
    def test_voltage_battery(self):
        data = bytes([1, LPP_VOLTAGE]) + struct.pack(">h", 380)
        result = parse_cayenne_lpp(data)
        assert len(result) == 1
        assert result[0]["channel"] == 1
        assert result[0]["name"] == "Voltage"
        assert abs(result[0]["value"] - 3.80) < 0.01
        assert result[0]["unit"] == "V"

    def test_battery_percent_from_voltage(self):
        pct = battery_percent_from_mv(3800)
        assert pct == 50

    def test_multi_channel_payload(self):
        data = b""
        data += bytes([1, LPP_VOLTAGE]) + struct.pack(">h", 420)
        data += bytes([2, LPP_TEMPERATURE]) + struct.pack(">h", 235)
        data += bytes([3, LPP_HUMIDITY, 120])
        result = parse_cayenne_lpp(data)
        assert len(result) == 3
        assert result[0]["channel"] == 1
        assert result[0]["type"] == LPP_VOLTAGE
        assert result[1]["channel"] == 2
        assert abs(result[1]["value"] - 23.5) < 0.01
        assert result[2]["channel"] == 3
        assert result[2]["value"] == 60.0

    def test_pressure(self):
        data = bytes([1, LPP_PRESSURE]) + struct.pack(">H", 10132)
        result = parse_cayenne_lpp(data)
        assert abs(result[0]["value"] - 1013.2) < 0.1
        assert result[0]["unit"] == "hPa"

    def test_luminosity(self):
        data = bytes([1, LPP_LUMINOSITY]) + struct.pack(">H", 450)
        result = parse_cayenne_lpp(data)
        assert result[0]["value"] == 450
        assert result[0]["unit"] == "lux"

    def test_digital_input(self):
        data = bytes([1, LPP_DIGITAL_INPUT, 1])
        result = parse_cayenne_lpp(data)
        assert result[0]["value"] == 1

    def test_switch(self):
        data = bytes([5, LPP_SWITCH, 0])
        result = parse_cayenne_lpp(data)
        assert result[0]["value"] == 0
        assert result[0]["name"] == "Switch"

    def test_percentage(self):
        data = bytes([1, LPP_PERCENTAGE, 75])
        result = parse_cayenne_lpp(data)
        assert result[0]["value"] == 75
        assert result[0]["unit"] == "%"

    def test_current(self):
        data = bytes([1, LPP_CURRENT]) + struct.pack(">h", 1500)
        result = parse_cayenne_lpp(data)
        assert abs(result[0]["value"] - 1.5) < 0.001
        assert result[0]["unit"] == "A"

    def test_frequency(self):
        data = bytes([1, LPP_FREQUENCY]) + struct.pack(">I", 868000000)
        result = parse_cayenne_lpp(data)
        assert result[0]["value"] == 868000000
        assert result[0]["unit"] == "Hz"

    def test_concentration_ppm(self):
        data = bytes([1, LPP_CONCENTRATION]) + struct.pack(">H", 400)
        result = parse_cayenne_lpp(data)
        assert result[0]["value"] == 400
        assert result[0]["unit"] == "ppm"

    def test_distance(self):
        data = bytes([1, LPP_DISTANCE]) + struct.pack(">I", 12500)
        result = parse_cayenne_lpp(data)
        assert abs(result[0]["value"] - 12.5) < 0.01
        assert result[0]["unit"] == "m"

    def test_unix_time(self):
        ts = 1719000000
        data = bytes([1, LPP_UNIX_TIME]) + struct.pack(">I", ts)
        result = parse_cayenne_lpp(data)
        assert result[0]["value"] == ts

    def test_accelerometer(self):
        data = bytes([1, LPP_ACCELEROMETER])
        data += struct.pack(">h", 100)  # x = 0.100G
        data += struct.pack(">h", -50)  # y = -0.050G
        data += struct.pack(">h", 9810)  # z = 9.810G
        result = parse_cayenne_lpp(data)
        val = result[0]["value"]
        assert abs(val["x"] - 0.100) < 0.001
        assert abs(val["y"] + 0.050) < 0.001
        assert abs(val["z"] - 9.810) < 0.001

    def test_colour_rgb(self):
        data = bytes([1, LPP_COLOUR, 255, 128, 0])
        result = parse_cayenne_lpp(data)
        val = result[0]["value"]
        assert val == {"r": 255, "g": 128, "b": 0}

    def test_gps_with_altitude(self):
        lat_raw = int(49.1953 * 10000).to_bytes(3, "big", signed=True)
        lon_raw = int(16.6068 * 10000).to_bytes(3, "big", signed=True)
        alt_raw = int(237.0 * 100).to_bytes(3, "big", signed=True)
        data = bytes([1, LPP_GPS]) + lat_raw + lon_raw + alt_raw
        result = parse_cayenne_lpp(data)
        val = result[0]["value"]
        assert abs(val["lat"] - 49.1953) < 0.001
        assert abs(val["lon"] - 16.6068) < 0.001
        assert abs(val["alt"] - 237.0) < 0.1

    def test_negative_temperature(self):
        data = bytes([1, LPP_TEMPERATURE]) + struct.pack(">h", -50)
        result = parse_cayenne_lpp(data)
        assert abs(result[0]["value"] - (-5.0)) < 0.01

    def test_full_sensor_payload(self):
        """Realistic sensor payload: battery + temp + humidity + pressure + GPS."""
        data = b""
        data += bytes([1, LPP_VOLTAGE]) + struct.pack(">h", 395)
        data += bytes([2, LPP_TEMPERATURE]) + struct.pack(">h", 218)
        data += bytes([3, LPP_HUMIDITY, 130])
        data += bytes([4, LPP_PRESSURE]) + struct.pack(">H", 10153)
        lat = int(50.0875 * 10000).to_bytes(3, "big", signed=True)
        lon = int(14.4213 * 10000).to_bytes(3, "big", signed=True)
        alt = int(195.0 * 100).to_bytes(3, "big", signed=True)
        data += bytes([5, LPP_GPS]) + lat + lon + alt
        result = parse_cayenne_lpp(data)
        assert len(result) == 5
        assert result[0]["type"] == LPP_VOLTAGE
        assert abs(result[0]["value"] - 3.95) < 0.01
        assert result[1]["type"] == LPP_TEMPERATURE
        assert abs(result[1]["value"] - 21.8) < 0.1
        assert result[2]["type"] == LPP_HUMIDITY
        assert result[2]["value"] == 65.0
        assert result[3]["type"] == LPP_PRESSURE
        assert abs(result[3]["value"] - 1015.3) < 0.1
        gps = result[4]["value"]
        assert abs(gps["lat"] - 50.0875) < 0.001
        assert abs(gps["lon"] - 14.4213) < 0.001

    def test_truncated_payload(self):
        data = bytes([1, LPP_TEMPERATURE, 0x01])  # missing second byte
        result = parse_cayenne_lpp(data)
        assert result == []

    def test_unknown_type_stops_parsing(self):
        data = bytes([1, 0xFE, 0x00, 0x00])  # unknown type
        result = parse_cayenne_lpp(data)
        assert result == []

    def test_same_channel_multiple_readings(self):
        data = b""
        data += bytes([1, LPP_TEMPERATURE]) + struct.pack(">h", 200)
        data += bytes([1, LPP_HUMIDITY, 100])
        result = parse_cayenne_lpp(data)
        assert len(result) == 2
        assert result[0]["channel"] == result[1]["channel"] == 1


# ─── Storage: sensor contacts ────────────────────────────────


class TestStorageSensorContacts:
    def test_save_and_load_sensor(self, storage):
        c = Contact(
            public_key=bytes(range(32)),
            name="Outdoor Sensor",
            type=ContactType.SENSOR,
        )
        storage.save_contact(c)
        loaded = storage.get_contacts()
        assert len(loaded) == 1
        assert loaded[0].name == "Outdoor Sensor"
        assert loaded[0].type == ContactType.SENSOR

    def test_sensor_type_persisted(self, storage):
        c = Contact(
            public_key=bytes([0xAA] * 32),
            name="Indoor",
            type=ContactType.SENSOR,
        )
        storage.save_contact(c)
        loaded = storage.get_contacts()
        assert loaded[0].type == 4
        assert loaded[0].type == ContactType.SENSOR

    def test_sensor_alongside_other_types(self, storage):
        chat = Contact(public_key=bytes([1] * 32), name="Alice", type=ContactType.CHAT)
        repeater = Contact(public_key=bytes([2] * 32), name="Relay-1", type=ContactType.REPEATER)
        sensor = Contact(public_key=bytes([3] * 32), name="Weather", type=ContactType.SENSOR)
        for c in (chat, repeater, sensor):
            storage.save_contact(c)
        loaded = storage.get_contacts()
        types = {c.name: c.type for c in loaded}
        assert types["Alice"] == ContactType.CHAT
        assert types["Relay-1"] == ContactType.REPEATER
        assert types["Weather"] == ContactType.SENSOR

    def test_sensor_update_preserves_type(self, storage):
        key = bytes([0xCC] * 32)
        c = Contact(public_key=key, name="Sensor-Old", type=ContactType.SENSOR)
        storage.save_contact(c)
        c2 = Contact(public_key=key, name="Sensor-New", type=ContactType.SENSOR)
        storage.save_contact(c2)
        loaded = storage.get_contacts()
        assert len(loaded) == 1
        assert loaded[0].name == "Sensor-New"
        assert loaded[0].type == ContactType.SENSOR

    def test_sensor_with_location(self, storage):
        c = Contact(
            public_key=bytes([0xDD] * 32),
            name="GPS Sensor",
            type=ContactType.SENSOR,
            latitude=50.0875,
            longitude=14.4213,
        )
        storage.save_contact(c)
        loaded = storage.get_contacts()
        assert loaded[0].has_location is True
        assert abs(loaded[0].latitude - 50.0875) < 0.001



"""Tests for app-side trip time measurement and ACK hash collision handling."""

import builtins
import gettext
import time
from unittest.mock import MagicMock, patch

if not hasattr(builtins, "_"):
    gettext.install("meshy")

from meshy.frame_handler import FrameHandler
from meshy.message_controller import MessageController
from meshy.models import Contact, ContactType, MessageStatus
from meshy.protocol import PUSH_CODE_SEND_CONFIRMED
from meshy.storage import Storage


def _make_contact(path_length=-1, path=b""):
    return Contact(
        public_key=bytes(range(32)),
        name="Test",
        type=ContactType.CHAT,
        path_length=path_length,
        path=path,
    )


def _make_ctrl():
    storage = Storage(db_path=":memory:")
    win = MagicMock()
    win._storage = storage
    win._contacts_by_key = {}
    ctrl = MessageController(win)
    return ctrl, storage


# ─── _do_send_message records send time ──────────────────────


class TestSendTimeRecording:
    def setup_method(self):
        self.ctrl, self.storage = _make_ctrl()

    def teardown_method(self):
        self.storage.close()

    def test_do_send_message_records_last_send_mono(self):
        contact = _make_contact()
        msg_id = "msg-1"
        self.ctrl.retry_context[msg_id] = {
            "contact": contact,
            "text": "hello",
            "timestamp": int(time.time()),
            "attempt": 0,
            "strategy": "force_flood",
        }
        before = time.monotonic()
        self.ctrl._do_send_message(msg_id)
        after = time.monotonic()

        ctx = self.ctrl.retry_context[msg_id]
        assert "last_send_mono" in ctx
        assert before <= ctx["last_send_mono"] <= after


# ─── register_ack_key stores send time ───────────────────────


class TestRegisterAckKey:
    def setup_method(self):
        self.ctrl, self.storage = _make_ctrl()

    def teardown_method(self):
        self.storage.close()

    def test_register_ack_key_stores_send_time(self):
        msg_id = "msg-1"
        send_time = time.monotonic()
        self.ctrl.retry_context[msg_id] = {
            "attempt": 0,
            "attempt_paths": {0: {"path_length": 2, "path_bytes": "aabb"}},
            "last_send_mono": send_time,
        }
        self.ctrl.register_ack_key(msg_id, "deadbeef")

        ctx = self.ctrl.retry_context[msg_id]
        assert ctx["ack_send_times"]["deadbeef"] == send_time
        assert ctx["ack_paths"]["deadbeef"] == {"path_length": 2, "path_bytes": "aabb"}

    def test_collision_overwrites_with_latest_send_time(self):
        msg_id = "msg-1"
        t0 = time.monotonic()
        self.ctrl.retry_context[msg_id] = {
            "attempt": 0,
            "attempt_paths": {
                0: {"path_length": 2, "path_bytes": "aabb"},
            },
            "last_send_mono": t0,
        }
        self.ctrl.register_ack_key(msg_id, "deadbeef")

        t4 = time.monotonic()
        ctx = self.ctrl.retry_context[msg_id]
        ctx["attempt"] = 4
        ctx["attempt_paths"][4] = {"path_length": -1, "path_bytes": ""}
        ctx["last_send_mono"] = t4
        self.ctrl.register_ack_key(msg_id, "deadbeef")

        assert ctx["ack_send_times"]["deadbeef"] == t4
        assert ctx["ack_paths"]["deadbeef"]["path_length"] == -1


# ─── Grace context preserves send times ──────────────────────


class TestGraceContextSendTimes:
    def setup_method(self):
        self.ctrl, self.storage = _make_ctrl()

    def teardown_method(self):
        self.storage.close()

    @patch("meshy.message_controller.GLib")
    def test_grace_context_includes_ack_send_times(self, mock_glib):
        contact = _make_contact()
        msg_id = "msg-1"
        send_time = 12345.678
        self.ctrl.retry_context[msg_id] = {
            "contact": contact,
            "attempt": 2,
            "strategy": "force_flood",
            "ack_paths": {"aabb": {"path_length": -1}},
            "ack_send_times": {"aabb": send_time},
            "path_used": {"path_length": -1},
        }
        self.ctrl._acks_by_msg[msg_id] = {"aabb"}
        self.ctrl.pending_acks["aabb"] = msg_id
        self.ctrl._on_message_timeout(msg_id, "aabb")

        assert msg_id in self.ctrl.grace_contexts
        assert self.ctrl.grace_contexts[msg_id]["ack_send_times"]["aabb"] == send_time


# ─── Frame handler uses app-side trip time ───────────────────


def _make_fh():
    storage = Storage(db_path=":memory:")
    win = MagicMock()
    win._storage = storage
    win._self_info = {"name": "Me"}
    win._channels_view.on_path_variant_received.return_value = False
    win._channels = []
    win._channels_view_inst = None
    win._conn_ctrl.channel_sync_index = -1
    win._device_info.path_hash_mode = 0
    ctrl = MessageController(win)
    win._msg_ctrl = ctrl
    fh = FrameHandler(win)
    return fh, ctrl, storage


def _ack_frame(ack_key_hex, trip_time_ms):
    return (
        bytes([PUSH_CODE_SEND_CONFIRMED])
        + bytes.fromhex(ack_key_hex)
        + trip_time_ms.to_bytes(4, "little")
    )


class TestFrameHandlerAppTripTime:
    def setup_method(self):
        self.fh, self.ctrl, self.storage = _make_fh()

    def teardown_method(self):
        self.storage.close()

    def test_primary_ack_uses_app_side_trip_time(self):
        msg_id = "msg-1"
        ack_key = "deadbeef"
        send_mono = time.monotonic() - 2.5

        contact = _make_contact()
        self.ctrl.pending_acks[ack_key] = msg_id
        self.ctrl._acks_by_msg[msg_id] = {ack_key}
        self.ctrl.retry_context[msg_id] = {
            "contact": contact,
            "attempt": 0,
            "strategy": "force_flood",
            "path_used": {"path_length": -1, "path_bytes": "", "hop_count": 0},
            "ack_paths": {ack_key: {"path_length": -1, "path_bytes": ""}},
            "ack_send_times": {ack_key: send_mono},
        }

        self.fh.handle_frame(_ack_frame(ack_key, 99999))

        status_call = self.fh._win.set_message_status.call_args
        assert status_call[0][1] == MessageStatus.DELIVERED
        detail = status_call[0][2]
        # App-side ~2500ms, not firmware's 99999ms
        assert "99" not in detail

    def test_fallback_to_firmware_trip_time_without_send_time(self):
        msg_id = "msg-2"
        ack_key = "cafebabe"
        contact = _make_contact()
        self.ctrl.pending_acks[ack_key] = msg_id
        self.ctrl._acks_by_msg[msg_id] = {ack_key}
        self.ctrl.retry_context[msg_id] = {
            "contact": contact,
            "attempt": 0,
            "strategy": "force_flood",
            "path_used": {"path_length": -1, "path_bytes": "", "hop_count": 0},
        }

        self.fh.handle_frame(_ack_frame(ack_key, 5000))

        status_call = self.fh._win.set_message_status.call_args
        detail = status_call[0][2]
        assert "5.0" in detail and "s" in detail

    def test_grace_ack_uses_app_side_trip_time(self):
        msg_id = "msg-3"
        ack_key = "11223344"
        send_mono = time.monotonic() - 1.2

        self.ctrl.grace_acks[ack_key] = msg_id
        self.ctrl._grace_by_msg[msg_id] = {ack_key}
        self.ctrl.grace_contexts[msg_id] = {
            "contact_key": "00" * 32,
            "ack_paths": {ack_key: {"path_length": 2, "path_bytes": "aabb", "hop_count": 2}},
            "ack_send_times": {ack_key: send_mono},
            "path_used": {"path_length": 2, "path_bytes": "aabb", "hop_count": 2},
        }

        self.fh.handle_frame(_ack_frame(ack_key, 88888))

        status_call = self.fh._win.set_message_status.call_args
        detail = status_call[0][2]
        assert "late" in detail
        # App-side ~1200ms, not firmware's 88888ms
        assert "88" not in detail

    def test_collision_scenario_uses_latest_attempt_time(self):
        """Attempts 0 and 4 produce the same ACK hash (attempt & 3 collision).

        Attempt 0 was sent ~100s ago (timed out), attempt 4 was sent ~2s ago.
        The ACK arrives and should report ~2s, not ~100s.
        """
        msg_id = "msg-4"
        ack_key = "aabbccdd"
        now = time.monotonic()
        t4 = now - 2  # attempt 4: 2s ago (attempt 0's time was overwritten)

        contact = _make_contact()
        self.ctrl.pending_acks[ack_key] = msg_id
        self.ctrl._acks_by_msg[msg_id] = {ack_key}
        self.ctrl.retry_context[msg_id] = {
            "contact": contact,
            "attempt": 4,
            "strategy": "smart",
            "path_used": {"path_length": -1, "path_bytes": "", "hop_count": 0},
            "ack_paths": {ack_key: {"path_length": -1, "path_bytes": ""}},
            "ack_send_times": {ack_key: t4},
        }

        self.fh.handle_frame(_ack_frame(ack_key, 100000))

        status_call = self.fh._win.set_message_status.call_args
        detail = status_call[0][2]
        trip_val = float(detail.replace(" ", "").replace(" ", "").rstrip("s").replace(",", "."))
        assert trip_val < 10  # app-side ~2s, not firmware's 100s



"""Tests for channel sender blocking and ghost message filtering in FrameHandler."""

import time
from unittest.mock import MagicMock

from meshy.frame_handler import FrameHandler
from meshy.models import Channel, NotificationLevel


def _make_channel(index=0, name="General", psk=None):
    return Channel(
        index=index,
        name=name,
        psk=psk or bytes(range(16)),
        notification_level=NotificationLevel.NONE,
    )


def _make_win(storage):
    win = MagicMock()
    win._storage = storage
    win._self_info = {"name": "Me"}
    win._channels_view.on_path_variant_received.return_value = False
    win._channels = [_make_channel()]
    win._channels_view_inst = None
    win._conn_ctrl.channel_sync_index = -1
    return win


def _make_msg(sender_name, channel_idx=0, text="hello"):
    return {
        "sender_name": sender_name,
        "msg_text": text,
        "sender_timestamp": time.time(),
        "channel_idx": channel_idx,
    }


class TestBlockedSenderFiltering:
    def test_blocked_message_is_dropped(self, storage):
        win = _make_win(storage)
        fh = FrameHandler(win)
        storage.block_channel_sender("Spammer")
        fh.load_blocked_senders()

        storage.save_channel_message = MagicMock()
        fh.handle_incoming_channel_message(_make_msg("Spammer"))

        storage.save_channel_message.assert_not_called()
        assert not win._channels_view.on_channel_message_received.called

    def test_unblocked_message_is_saved(self, storage):
        win = _make_win(storage)
        fh = FrameHandler(win)
        fh.load_blocked_senders()

        fh.handle_incoming_channel_message(_make_msg("Alice"))

        win._channels_view.on_channel_message_received.assert_called_once()

    def test_update_blocked_senders_reflects_changes(self, storage):
        win = _make_win(storage)
        fh = FrameHandler(win)
        fh.load_blocked_senders()
        assert fh._blocked_senders == set()

        storage.block_channel_sender("Spammer")
        fh.update_blocked_senders()
        assert "Spammer" in fh._blocked_senders

        storage.unblock_channel_sender("Spammer")
        fh.update_blocked_senders()
        assert "Spammer" not in fh._blocked_senders

    def test_echo_not_affected_by_block(self, storage):
        win = _make_win(storage)
        fh = FrameHandler(win)
        storage.block_channel_sender("Me")
        fh.load_blocked_senders()

        fh.handle_incoming_channel_message(_make_msg("Me"))

        win._channels_view.on_repeat_received.assert_called_once()
        win._channels_view.on_channel_message_received.assert_not_called()

    def test_empty_sender_name_not_blocked(self, storage):
        win = _make_win(storage)
        fh = FrameHandler(win)
        storage.block_channel_sender("")
        fh.load_blocked_senders()

        fh.handle_incoming_channel_message(_make_msg(""))

        win._channels_view.on_channel_message_received.assert_called_once()


class TestGhostMessageFiltering:
    """Guard against firmware delivering messages for unconfigured channels.

    Empty channel slots (all-zero PSK) get hash 0x37 in the firmware and can
    receive crafted group messages.  After channel sync completes Meshy must
    drop these.
    """

    def test_message_for_unconfigured_channel_dropped(self, storage):
        win = _make_win(storage)
        fh = FrameHandler(win)
        fh.load_blocked_senders()

        fh.handle_incoming_channel_message(_make_msg("Alice", channel_idx=5))

        win._channels_view.on_channel_message_received.assert_not_called()

    def test_message_for_configured_channel_accepted(self, storage):
        win = _make_win(storage)
        fh = FrameHandler(win)
        fh.load_blocked_senders()

        fh.handle_incoming_channel_message(_make_msg("Alice", channel_idx=0))

        win._channels_view.on_channel_message_received.assert_called_once()

    def test_empty_channel_slot_treated_as_unconfigured(self, storage):
        """A slot with empty name and zero PSK must not accept messages."""
        win = _make_win(storage)
        ghost = Channel.empty(3)
        win._channels.append(ghost)
        fh = FrameHandler(win)
        fh.load_blocked_senders()

        fh.handle_incoming_channel_message(_make_msg("Alice", channel_idx=3))

        win._channels_view.on_channel_message_received.assert_not_called()

    def test_guard_inactive_during_channel_sync(self, storage):
        """During sync, messages pass through even for unknown indices."""
        win = _make_win(storage)
        win._conn_ctrl.channel_sync_index = 2  # sync in progress
        fh = FrameHandler(win)
        fh.load_blocked_senders()

        fh.handle_incoming_channel_message(_make_msg("Alice", channel_idx=5))

        win._channels_view.on_channel_message_received.assert_called_once()



# Copyright 2026, Jiri Eischmann <jiri@eischmann.cz> and the meshy contributors
# SPDX-License-Identifier: GPL-3.0-or-later

"""Pure logic extracted from RepeaterView for testability without GTK."""

ACL_ROLES = {
    3: "Admin",
    2: "Read-Write",
    1: "Read Only",
}

ROLE_PERMS = {0: 3, 1: 2, 2: 1}


def parse_acl_response(data):
    """Parse raw ACL binary response into list of {'key': hex, 'perm': int}.

    Response format: [code(1)][reserved(1)][tag(4)][payload...]
    Payload: 7-byte entries: [pubkey_prefix(6)][perm(1)]
    """
    buf = data[6:]
    entries = []
    i = 0
    while i + 7 <= len(buf):
        key = buf[i : i + 6].hex()
        perm = buf[i + 6]
        if key != "000000000000":
            entries.append({"key": key, "perm": perm})
        i += 7
    return entries


def parse_region_text(region_text, default_text=""):
    """Parse region tree text and default scope into structured data.

    Returns (entries, default_scope_name) where each entry is:
        {"name": str, "indent": int, "home": bool, "flood": bool}
    """
    default_name = ""
    if "default scope is" in default_text:
        val = default_text.split("default scope is", 1)[1].strip()
        if val and val != "<null>":
            default_name = val

    entries = []
    for line in region_text.split("\n"):
        if not line.strip():
            continue
        indent = len(line) - len(line.lstrip(" "))
        stripped = line.strip()
        flood = stripped.endswith(" F")
        if flood:
            stripped = stripped[:-2].strip()
        home = stripped.endswith("^")
        if home:
            stripped = stripped[:-1]
        entries.append(
            {
                "name": stripped,
                "indent": indent,
                "home": home,
                "flood": flood,
            }
        )
    return entries, default_name



"""Tests for contact sorting and last_message_at preservation."""

import builtins
import gettext
import importlib.util
import sys
from datetime import datetime, timedelta
from pathlib import Path
from unittest.mock import MagicMock

if not hasattr(builtins, "_"):
    gettext.install("meshy")

from meshy.models import Contact, ContactType
from meshy.storage import Storage

# contacts_view.py contains pure-Python sort logic but its package __init__
# pulls in GTK/Shumate.  Load the single file directly to skip that chain.
_src = Path(__file__).resolve().parent.parent / "src" / "views" / "contacts_view.py"
_spec = importlib.util.spec_from_file_location("_contacts_view_isolated", _src)
_mod = importlib.util.module_from_spec(_spec)

_GI_STUBS = ["gi", "gi.repository"]
_gi_originals = {m: sys.modules.pop(m, None) for m in _GI_STUBS}
for m in _GI_STUBS:
    sys.modules[m] = MagicMock()
_spec.loader.exec_module(_mod)
for m in _GI_STUBS:
    sys.modules.pop(m, None)
    if _gi_originals[m] is not None:
        sys.modules[m] = _gi_originals[m]

contact_sort_key = _mod.contact_sort_key
SORT_AZ = _mod.SORT_AZ
SORT_HEARD = _mod.SORT_HEARD
SORT_MESSAGES = _mod.SORT_MESSAGES


def _contact(name, *, fav=False, last_msg=None, last_seen=None, ctype=ContactType.CHAT):
    return Contact(
        public_key=bytes(range(32)),
        name=name,
        type=ctype,
        flags=1 if fav else 0,
        last_message_at=last_msg,
        last_seen=last_seen,
    )


# ─── Sort by name (A-Z) ────────────────────────────────────


class TestSortAZ:
    def test_alphabetical_order(self):
        contacts = [_contact("Charlie"), _contact("Alice"), _contact("Bob")]
        contacts.sort(key=lambda c: contact_sort_key(c, SORT_AZ, False))
        assert [c.name for c in contacts] == ["Alice", "Bob", "Charlie"]

    def test_case_insensitive(self):
        contacts = [_contact("bob"), _contact("Alice")]
        contacts.sort(key=lambda c: contact_sort_key(c, SORT_AZ, False))
        assert [c.name for c in contacts] == ["Alice", "bob"]

    def test_favorites_first(self):
        contacts = [
            _contact("Zara", fav=False),
            _contact("Alice", fav=True),
            _contact("Bob", fav=False),
        ]
        contacts.sort(key=lambda c: contact_sort_key(c, SORT_AZ, True))
        assert [c.name for c in contacts] == ["Alice", "Bob", "Zara"]

    def test_favorites_sorted_among_themselves(self):
        contacts = [
            _contact("Zara", fav=True),
            _contact("Alice", fav=True),
            _contact("Bob", fav=False),
        ]
        contacts.sort(key=lambda c: contact_sort_key(c, SORT_AZ, True))
        assert [c.name for c in contacts] == ["Alice", "Zara", "Bob"]


# ─── Sort by latest message ────────────────────────────────


class TestSortMessages:
    def test_newest_message_first(self):
        now = datetime.now()
        contacts = [
            _contact("Old", last_msg=now - timedelta(hours=2)),
            _contact("New", last_msg=now),
            _contact("Mid", last_msg=now - timedelta(hours=1)),
        ]
        contacts.sort(key=lambda c: contact_sort_key(c, SORT_MESSAGES, False))
        assert [c.name for c in contacts] == ["New", "Mid", "Old"]

    def test_no_message_sorts_last(self):
        now = datetime.now()
        contacts = [
            _contact("NoMsg"),
            _contact("HasMsg", last_msg=now),
        ]
        contacts.sort(key=lambda c: contact_sort_key(c, SORT_MESSAGES, False))
        assert [c.name for c in contacts] == ["HasMsg", "NoMsg"]

    def test_favorites_first_with_messages(self):
        now = datetime.now()
        contacts = [
            _contact("NonFav", last_msg=now),
            _contact("Fav", fav=True, last_msg=now - timedelta(hours=5)),
        ]
        contacts.sort(key=lambda c: contact_sort_key(c, SORT_MESSAGES, True))
        assert [c.name for c in contacts] == ["Fav", "NonFav"]

    def test_repeater_without_message_below_chat_with_message(self):
        now = datetime.now()
        contacts = [
            _contact("Repeater", fav=True, ctype=ContactType.REPEATER),
            _contact("ChatContact", fav=True, last_msg=now),
        ]
        contacts.sort(key=lambda c: contact_sort_key(c, SORT_MESSAGES, True))
        assert [c.name for c in contacts] == ["ChatContact", "Repeater"]


# ─── Sort by last heard ─────────────────────────────────────


class TestSortHeard:
    def test_most_recently_heard_first(self):
        now = datetime.now()
        contacts = [
            _contact("Old", last_seen=now - timedelta(hours=3)),
            _contact("New", last_seen=now),
        ]
        contacts.sort(key=lambda c: contact_sort_key(c, SORT_HEARD, False))
        assert [c.name for c in contacts] == ["New", "Old"]

    def test_no_last_seen_sorts_last(self):
        now = datetime.now()
        contacts = [
            _contact("NeverSeen"),
            _contact("Seen", last_seen=now),
        ]
        contacts.sort(key=lambda c: contact_sort_key(c, SORT_HEARD, False))
        assert [c.name for c in contacts] == ["Seen", "NeverSeen"]


# ─── last_message_at preservation (storage) ─────────────────


class TestStorageLastMessageAtPreservation:
    def setup_method(self):
        self.storage = Storage(db_path=":memory:")

    def teardown_method(self):
        self.storage.close()

    def test_save_contact_preserves_last_message_at_via_coalesce(self):
        now = datetime.now()
        c = Contact(
            public_key=bytes(range(32)),
            name="Alice",
            type=ContactType.CHAT,
            last_message_at=now,
        )
        self.storage.save_contact(c)

        c_update = Contact(
            public_key=bytes(range(32)),
            name="Alice",
            type=ContactType.CHAT,
            last_message_at=None,
        )
        self.storage.save_contact(c_update)

        loaded = self.storage.get_contacts()
        alice = [x for x in loaded if x.name == "Alice"][0]
        assert alice.last_message_at is not None
        assert abs(alice.last_message_at.timestamp() - now.timestamp()) < 1

    def test_save_contact_updates_last_message_at_when_provided(self):
        old = datetime(2025, 1, 1)
        new = datetime(2026, 6, 15)
        c = Contact(
            public_key=bytes(range(32)),
            name="Bob",
            type=ContactType.CHAT,
            last_message_at=old,
        )
        self.storage.save_contact(c)

        c.last_message_at = new
        self.storage.save_contact(c)

        loaded = self.storage.get_contacts()
        bob = [x for x in loaded if x.name == "Bob"][0]
        assert abs(bob.last_message_at.timestamp() - new.timestamp()) < 1



"""Tests for the shared format_trip_time utility."""

import locale

from meshy.utils import format_trip_time

NNBSP = " "


class TestFormatTripTime:
    def test_none_returns_empty(self):
        assert format_trip_time(None) == ""

    def test_zero_returns_empty(self):
        assert format_trip_time(0) == ""

    def test_sub_second(self):
        result = format_trip_time(450)
        assert result.endswith(NNBSP + "s")
        num = result.removesuffix(NNBSP + "s")
        assert float(num.replace(",", ".")) == 0.4 or float(num.replace(",", ".")) == 0.5

    def test_one_second(self):
        result = format_trip_time(1000)
        num = result.removesuffix(NNBSP + "s")
        assert float(num.replace(",", ".")) == 1.0

    def test_multi_second(self):
        result = format_trip_time(2500)
        num = result.removesuffix(NNBSP + "s")
        assert float(num.replace(",", ".")) == 2.5

    def test_large_value(self):
        result = format_trip_time(15000)
        num = result.removesuffix(NNBSP + "s")
        assert float(num.replace(",", ".")) == 15.0

    def test_small_value(self):
        result = format_trip_time(50)
        assert result.endswith(NNBSP + "s")
        num = result.removesuffix(NNBSP + "s")
        assert float(num.replace(",", ".")) == 0.1 or float(num.replace(",", ".")) == 0.0

    def test_narrow_no_break_space_before_unit(self):
        result = format_trip_time(1000)
        assert NNBSP + "s" in result

    def test_one_decimal_place(self):
        result = format_trip_time(3000)
        num = result.removesuffix(NNBSP + "s")
        sep = locale.localeconv()["decimal_point"]
        parts = num.split(sep)
        assert len(parts) == 2
        assert len(parts[1]) == 1
